#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import time

import eventlet
from oslo_log import log as logging
import paramiko
import yaml

from tacker.common import cmd_executer
from tacker.common import exceptions
from tacker.common import log
from tacker.vnflcm import utils as vnflcm_utils
from tacker.vnfm.mgmt_drivers import vnflcm_abstract_driver

LOG = logging.getLogger(__name__)
FREE5GC_CMD_TIMEOUT = 30


class Free5gcMgmtDriverCnf(vnflcm_abstract_driver.VnflcmMgmtAbstractDriver):
    def get_type(self):
        return 'mgmt-drivers-free5gc-cnf'

    def get_name(self):
        return 'mgmt-drivers-free5gc-cnf'

    def get_description(self):
        return 'Tacker VNFMgmt Free5gc CNF Driver'

    @log.log
    def instantiate_start(self, context, vnf_instance,
                          instantiate_vnf_request, grant,
                          grant_request, **kwargs):
        vnf_package_path = vnflcm_utils._get_vnf_package_path(
            context, vnf_instance.vnfd_id)
        kubernetes_file_paths = instantiate_vnf_request.\
            additional_params.get('lcm-kubernetes-def-files')
        for path in kubernetes_file_paths:
            if 'configmap' in path:
                configmap_path = os.path.join(vnf_package_path, path)
            if 'amf' in path:
                amf_path = os.path.join(vnf_package_path, path)
            if 'smf' in path:
                smf_path = os.path.join(vnf_package_path, path)
            if 'upf' in path:
                upf_path = os.path.join(vnf_package_path, path)
        with open(amf_path) as f:
            results = yaml.safe_load_all(f)
            for result in results:
                if result.get('kind') == 'Deployment':
                    amf_init_containers = result.get('spec').get(
                        'template').get('spec').get('initContainers')
                    amf_ip_list = []
                    for amf_init_container in amf_init_containers:
                        amf_ip_args = amf_init_container.get('args')
                        for amf_ip_arg in amf_ip_args:
                            if '-i=' in amf_ip_arg:
                                amf_ip = amf_ip_arg.replace('-i=', '')
                                amf_ip = amf_ip.partition('/')[0]
                                amf_ip_list.append(amf_ip)
        with open(smf_path) as f:
            results = yaml.safe_load_all(f)
            for result in results:
                if result.get('kind') == 'Deployment':
                    smf_init_containers = result.get('spec').get(
                        'template').get('spec').get('initContainers')
                    smf_ip_list = []
                    for smf_init_container in smf_init_containers:
                        smf_ip_args = smf_init_container.get('args')
                        for smf_ip_arg in smf_ip_args:
                            if '-i=' in smf_ip_arg:
                                smf_ip = smf_ip_arg.replace('-i=', '')
                                smf_ip = smf_ip.partition('/')[0]
                                smf_ip_list.append(smf_ip)
        with open(upf_path) as f:
            results = yaml.safe_load_all(f)
            for result in results:
                if result.get('kind') == 'Deployment':
                    upf_init_containers = result.get('spec').get(
                        'template').get('spec').get('initContainers')
                    upf_ip_list = []
                    for upf_init_container in upf_init_containers:
                        upf_ip_args = upf_init_container.get('args')
                        for upf_ip_arg in upf_ip_args:
                            if '-i=' in upf_ip_arg:
                                upf_ip = upf_ip_arg.replace('-i=', '')
                                upf_ip = upf_ip.partition('/')[0]
                                upf_ip_list.append(upf_ip)

        with open(configmap_path, encoding='utf-8') as f:
            results = yaml.safe_load_all(f)
            for result in results:
                # check amfcfg.yaml in configmap
                amf_file = result.get('data').get('amfcfg.yaml')
                index_start = \
                    amf_file.index('ngapIpList') + len('ngapIpList') + 1
                index_end = amf_file.index('sbi')
                amf_ip_str = amf_file[index_start:index_end]
                count = 0
                for amf_ip in amf_ip_list:
                    if amf_ip in amf_ip_str:
                        count = count + 1
                if count == 0:
                    LOG.error('The configmap of amfcfg.yaml is invalid.'
                              ' "ngapIpList" may be wrong.')
                    raise exceptions.MgmtDriverOtherError(
                        'The configmap of amfcfg.yaml is invalid.'
                        ' "ngapIpList" may be wrong.')

                # check smfcfg.yaml in configmap
                smf_file = result.get('data').get('smfcfg.yaml')
                index_start = smf_file.index('pfcp') + len('pfcp') + 1
                index_end = smf_file.index('userplane_information')
                smf_ip_str = smf_file[index_start:index_end]
                for smf_pfcp in smf_ip_list:
                    if smf_pfcp in smf_ip_str:
                        break
                else:
                    LOG.error('The configmap of smfcfg.yaml is invalid.'
                              ' "pfcp" may be wrong.')
                    raise exceptions.MgmtDriverOtherError(
                        'The configmap of smfcfg.yaml is invalid.'
                        ' "pfcp" may be wrong.')
                index_start2 = smf_file.index('UPF:') + len('UPF:') + 1
                index_end2 = smf_file.index('sNssaiUpfInfos')
                upf_pfcp_ip_str = smf_file[index_start2:index_end2]
                for upf_ip in upf_ip_list:
                    if upf_ip in upf_pfcp_ip_str:
                        break
                else:
                    LOG.error('The configmap of smfcfg.yaml is invalid.'
                              ' The node_id of UPF may be wrong.')
                    raise exceptions.MgmtDriverOtherError(
                        'The configmap of smfcfg.yaml is invalid.'
                        ' The node_id of UPF may be wrong.')

    def _execute_command(self, commander, ssh_command, timeout, type, retry):
        eventlet.monkey_patch()
        while retry >= 0:
            try:
                with eventlet.Timeout(timeout, True):
                    result = commander.execute_command(
                        ssh_command, input_data=None)
                    break
            except eventlet.timeout.Timeout:
                LOG.debug('It is time out, When execute command: '
                          '{}.'.format(ssh_command))
                retry -= 1
                if retry < 0:
                    LOG.error('It is time out, When execute command: '
                              '{}.'.format(ssh_command))
                    raise exceptions.MgmtDriverOtherError(
                        error_message='It is time out, When execute command: '
                                      '{}.'.format(ssh_command))
                time.sleep(30)
        if type == 'common':
            err = result.get_stderr()
            if err:
                LOG.error(err)
                raise exceptions.MgmtDriverRemoteCommandError(err_info=err)
        return result.get_stdout()

    def _check_values(self, additional_param):
        if not additional_param.get('master_node_username'):
            LOG.error('The master_node_username in the '
                      'additionalParams cannot be None.')
            raise exceptions.MgmtDriverNotFound(
                param='master_node_username')
        if not additional_param.get('master_node_password'):
            LOG.error('The master_node_password in the '
                      'additionalParams cannot be None.')
            raise exceptions.MgmtDriverNotFound(
                param='master_node_username')
        if not additional_param.get('ssh_master_node_ip'):
            LOG.error('The ssh_master_node_ip in the '
                      'additionalParams cannot be None.')
            raise exceptions.MgmtDriverNotFound(
                param='ssh_master_node_ip')

    def _send_and_receive_file(self, host, user, password,
                               remote_file, local_file, operation):
        connect = paramiko.Transport(host, 22)
        connect.connect(username=user, password=password)
        sftp = paramiko.SFTPClient.from_transport(connect)
        if operation == 'receive':
            sftp.get(remote_file, local_file)
        else:
            sftp.put(local_file, remote_file)
        connect.close()

    @log.log
    def instantiate_end(self, context, vnf_instance,
                        instantiate_vnf_request, grant,
                        grant_request, **kwargs):
        additional_param = instantiate_vnf_request.\
            additional_params.get('free5gc', {})
        self._check_values(additional_param)
        ssh_master_node_ip = additional_param.get('ssh_master_node_ip')
        master_node_username = additional_param.get('master_node_username')
        master_node_password = additional_param.get('master_node_password')
        if not additional_param.get('upf_config_file_path'):
            upf_config_file_path = \
                '/go/src/free5gc/NFs/upf/build/config/upfcfg.yaml'
        if not additional_param.get('smf_config_file_path'):
            smf_config_file_path = '/go/src/free5gc/config/smfcfg.yaml'

        commander = cmd_executer.RemoteCommandExecutor(
            user=master_node_username, password=master_node_password,
            host=ssh_master_node_ip,
            timeout=30)
        # get upf ip from smf
        ssh_command = "kubectl get pod | grep smf | awk '{print $1}'"
        smf_pod_name = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)[0].replace('\n', '')
        ssh_command = ("kubectl cp {smf_pod_name}:{smf_config_file_path}"
                       " /tmp/smfcfg.yaml"
                       .format(smf_pod_name=smf_pod_name,
                               smf_config_file_path=smf_config_file_path))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        local_smf_path = '/tmp/smfcfg.yaml'
        self._send_and_receive_file(
            ssh_master_node_ip, master_node_username, master_node_password,
            '/tmp/smfcfg.yaml', local_smf_path, 'receive')
        upf_gtpu_list = []
        with open(local_smf_path) as f:
            file_content = yaml.safe_load(f)
            upf_pfcp_ip = file_content['configuration'][
                'userplane_information']['up_nodes']['UPF']['node_id']
            upf_gtpu_interface_list = file_content['configuration'][
                'userplane_information']['up_nodes']['UPF']['interfaces']
            for upf_gtpu_interface in upf_gtpu_interface_list:
                upf_gtpu_list = (upf_gtpu_interface['endpoints'] +
                                 upf_gtpu_list)

        # modify upf info
        upf_example_file_path = \
            '/go/src/free5gc/NFs/upf/build/config/upfcfg.yaml'
        ssh_command = "kubectl get pod | grep upf | awk '{print $1}'"
        upf_pod_name = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)[0].replace('\n', '')
        ssh_command = ("kubectl cp {upf_pod_name}:{upf_example_file_path}"
                       " /tmp/upfcfg.yaml -c myapp-container"
                       .format(upf_pod_name=upf_pod_name,
                               upf_example_file_path=upf_example_file_path))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        local_upf_path = '/tmp/upfcfg.yaml'
        self._send_and_receive_file(
            ssh_master_node_ip, master_node_username, master_node_password,
            '/tmp/upfcfg.yaml', local_upf_path, 'receive')
        with open(local_upf_path, 'r') as f:
            upf_content = yaml.safe_load(f)
            upf_content['configuration']['pfcp'][0]['addr'] = upf_pfcp_ip
            for index in range(len(upf_gtpu_list)):
                upf_content['configuration']['gtpu'][index]['addr'] =\
                    upf_gtpu_list[index]
        with open(local_upf_path, 'w') as nf:
            yaml.safe_dump(upf_content, nf, default_flow_style=False)
        self._send_and_receive_file(
            ssh_master_node_ip, master_node_username, master_node_password,
            '/tmp/upfcfg.yaml', local_upf_path, 'send')
        ssh_command = ("kubectl cp /tmp/upfcfg.yaml"
                       " {upf_pod_name}:{upf_config_file_path}"
                       " -c myapp-container"
                       .format(upf_pod_name=upf_pod_name,
                               upf_config_file_path=upf_config_file_path))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)

        # start upf process
        ssh_command = ('cat <<EOF | sudo tee "run_upf.sh" >/dev/null\n'
                       'kubectl exec'
                       ' {} -i -- sh'
                       '<< eof\n'
                       'ip link delete upfgtp\neof\n'
                       'kubectl exec'
                       ' {} -i -- sh'
                       '<< eof\n'
                       './NFs/upf/build/bin/free5gc-upfd -f {}\neof'
                       '\nEOF\n'
                       .format(upf_pod_name, upf_pod_name,
                               upf_config_file_path))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        ssh_command = "sudo chmod 777 run_upf.sh"
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        ssh_command = "nohup ./run_upf.sh > upf.txt  2>&1 &"
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)

        # restart smf
        ssh_command = "kubectl get pod {} -o yaml | " \
                      "kubectl replace --force -f -".format(smf_pod_name)
        self._execute_command(
            commander, ssh_command, 120,
            'common', 0)
        time.sleep(120)
        ssh_command = "kubectl get pod | grep smf | awk '{print $1}'"
        smf_pod_name = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)[0].replace('\n', '')
        ssh_command = "kubectl get pod {} | " \
                      "grep 'Running'".format(smf_pod_name)
        result = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 0)
        if not result:
            LOG.error('SMF restart failed. Please check'
                      ' you k8s-cluster environment.')
            raise exceptions.MgmtDriverOtherError(
                'SMF restart failed. Please check you'
                ' k8s-cluster environment.')
        commander.close_session()
        os.remove(local_upf_path)
        os.remove(local_smf_path)

    @log.log
    def terminate_start(self, context, vnf_instance,
                        terminate_vnf_request, grant,
                        grant_request, **kwargs):
        pass

    @log.log
    def terminate_end(self, context, vnf_instance,
                      terminate_vnf_request, grant,
                      grant_request, **kwargs):
        pass

    @log.log
    def scale_start(self, context, vnf_instance,
                    scale_vnf_request, grant,
                    grant_request, **kwargs):
        pass

    @log.log
    def scale_end(self, context, vnf_instance,
                  scale_vnf_request, grant,
                  grant_request, **kwargs):
        additional_param = vnf_instance.instantiated_vnf_info.\
            additional_params.get('free5gc', {})
        ssh_master_node_ip = additional_param.get('ssh_master_node_ip')
        master_node_username = additional_param.get('master_node_username')
        master_node_password = additional_param.get('master_node_password')
        if not additional_param.get('upf_config_file_path'):
            upf_config_file_path = \
                '/go/src/free5gc/NFs/upf/build/config/upfcfg.yaml'
        if not additional_param.get('smf_config_file_path'):
            smf_config_file_path = '/go/src/free5gc/config/smfcfg.yaml'

        commander = cmd_executer.RemoteCommandExecutor(
            user=master_node_username, password=master_node_password,
            host=ssh_master_node_ip,
            timeout=30)
        ssh_command = "kubectl get pod | grep smf | awk '{print $1}'"
        smf_pod_name = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)[0].replace('\n', '')
        # get upf2's ip from smfcfg.yaml
        ssh_command = ("kubectl cp {smf_pod_name}:{smf_config_file_path}"
                       " /tmp/smfcfg.yaml -c myapp-container"
                       .format(smf_pod_name=smf_pod_name,
                               smf_config_file_path=smf_config_file_path))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        local_smf_path = '/tmp/smfcfg.yaml'
        self._send_and_receive_file(
            ssh_master_node_ip, master_node_username, master_node_password,
            '/tmp/smfcfg.yaml', local_smf_path, 'receive')
        upf_gtpu_list = []
        with open(local_smf_path) as f:
            file_content = yaml.safe_load(f)
            upf_pfcp_ip = file_content['configuration'][
                'userplane_information']['up_nodes']['UPF2']['node_id']
            upf_gtpu_interface_list = file_content['configuration'][
                'userplane_information']['up_nodes']['UPF2']['interfaces']
            for upf_gtpu_interface in upf_gtpu_interface_list:
                upf_gtpu_list = (upf_gtpu_interface['endpoints'] +
                                 upf_gtpu_list)

        # modify upf2's config file
        ssh_command = "kubectl get pod | grep upf | awk '{print $5}'"
        upf_pod_age_list = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        count = 1
        age_all = {}
        age_day = 0
        age_h = 0
        age_m = 0
        age_s = 0
        for age in upf_pod_age_list:
            if 'd' in age:
                age_day = age.split('d')[0]
                if 'h' in age:
                    age_h = age.split('d')[1].split('h')[0]
            elif 'h' in age and not age_h:
                age_h = age.split('h')[0]
                if 'm' in age:
                    age_m = age.split('h')[1].split('m')[0]
            elif 'm' in age and not age_m:
                age_m = age.split('m')[0]
                if 's' in age:
                    age_s = age.split('m')[1].split('s')[0]
            elif 's' in age and not age_s:
                age_s = age.split('s')[0]
            age_all[count] = \
                int(age_day) * 24 * 60 * 60 + int(age_h) * 60 * 60 +\
                int(age_m) * 60 + int(age_s) * 60
            count = count + 1
        age1 = age_all[1]
        age2 = age_all[2]
        if age1 > age2:
            scale_count = 1
        else:
            scale_count = 0
        ssh_command = ("kubectl get pod | grep upf | grep Running | awk '{"
                       "print $1}'")
        upf_pod_name_list = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        upf2_pod_name = upf_pod_name_list[scale_count].replace('\n', '')
        upf_example_file_path = \
            '/go/src/free5gc/NFs/upf/build/config/upfcfg.yaml'
        ssh_command = ("kubectl cp {upf_pod_name}:{upf_example_file_path}"
                       " /tmp/upfcfg.yaml -c myapp-container"
                       .format(upf_pod_name=upf2_pod_name,
                               upf_example_file_path=upf_example_file_path))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        local_upf_path = '/tmp/upfcfg.yaml'
        self._send_and_receive_file(
            ssh_master_node_ip, master_node_username, master_node_password,
            '/tmp/upfcfg.yaml', local_upf_path, 'receive')
        with open(local_upf_path, 'r') as f:
            upf_content = yaml.safe_load(f)
            upf_content['configuration']['pfcp'][0]['addr'] = upf_pfcp_ip
            for index in range(len(upf_gtpu_list)):
                upf_content['configuration']['gtpu'][index]['addr'] = \
                    upf_gtpu_list[index]
        with open(local_upf_path, 'w') as nf:
            yaml.safe_dump(upf_content, nf, default_flow_style=False)
        self._send_and_receive_file(
            ssh_master_node_ip, master_node_username, master_node_password,
            '/tmp/upfcfg.yaml', local_upf_path, 'send')
        ssh_command = ("kubectl cp /tmp/upfcfg.yaml"
                       " {upf_pod_name}:{upf_config_file_path} "
                       "-c myapp-container"
                       .format(upf_pod_name=upf2_pod_name,
                               upf_config_file_path=upf_config_file_path))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        ssh_command = ("kubectl exec {upf2_pod_name} -c myapp-container -- "
                       "ifconfig eth1 {ip}/24"
                       .format(upf2_pod_name=upf2_pod_name,
                               ip=upf_gtpu_list[0]))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        ssh_command = ("kubectl exec {upf2_pod_name} -c myapp-container -- "
                       "ifconfig eth2 {ip}/23"
                       .format(upf2_pod_name=upf2_pod_name, ip=upf_pfcp_ip))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        ssh_command = ("kubectl exec {upf2_pod_name} -c myapp-container -- "
                       "ifconfig eth3 192.168.52.253/24"
                       .format(upf2_pod_name=upf2_pod_name))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        # start upf process
        ssh_command = ('cat <<EOF | sudo tee "run_upf2.sh" >/dev/null\n'
                       'kubectl exec'
                       ' {} -i -- sh'
                       '<< eof\n'
                       'ip link delete upfgtp\neof\n'
                       'kubectl exec'
                       ' {} -i -- sh'
                       '<< eof\n'
                       './NFs/upf/build/bin/free5gc-upfd -f {}\neof'
                       '\nEOF\n'.format(upf2_pod_name, upf2_pod_name,
                                        upf_config_file_path))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        ssh_command = "sudo chmod 777 run_upf2.sh"
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        ssh_command = "nohup ./run_upf2.sh > upf2.txt 2>&1 &"
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)

        # restrat smf process
        ssh_command = "kubectl get pod {} -o yaml | " \
                      "kubectl replace --force -f -".format(smf_pod_name)
        self._execute_command(
            commander, ssh_command, 120,
            'common', 0)
        time.sleep(120)
        ssh_command = "kubectl get pod | grep smf | awk '{print $1}'"
        smf_pod_name = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)[0].replace('\n', '')
        ssh_command = "kubectl get pod {} | " \
                      "grep 'Running'".format(smf_pod_name)
        result = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 0)
        if not result:
            LOG.error('SMF restart failed. Please check'
                      ' you k8s-cluster environment.')
            raise exceptions.MgmtDriverOtherError(
                'SMF restart failed. Please check you'
                ' k8s-cluster environment.')
        commander.close_session()
        os.remove(local_smf_path)
        os.remove(local_upf_path)

        # if pod-affinity rule exists, check the pod deployed on different
        # worker
        affinity_flag = False
        artifact_files = vnf_instance.instantiated_vnf_info.\
            additional_params.get('lcm-kubernetes-def-files', {})
        vnf_package_path = vnflcm_utils._get_vnf_package_path(
            context, vnf_instance.vnfd_id)
        for artifact_file in artifact_files:
            if 'upf' in artifact_file:
                upf_file_path = os.path.join(
                    vnf_package_path, artifact_file)
                LOG.debug('upf_path:{}'.format(upf_file_path))
                with open(upf_file_path) as f:
                    yaml_content_all = yaml.safe_load_all(f.read())
                for yaml_content in yaml_content_all:
                    if (yaml_content['spec']['template']['spec']
                            .get('affinity')):
                        affinity_rule = (yaml_content['spec']['template']
                                         ['spec'].get('affinity'))
                        if affinity_rule.get('podAntiAffinity'):
                            affinity_flag = True
        LOG.debug('affinity_flag:{}'.format(affinity_flag))
        if affinity_flag:
            commander = cmd_executer.RemoteCommandExecutor(
                user=master_node_username, password=master_node_password,
                host=ssh_master_node_ip,
                timeout=30)
            ssh_command = ("kubectl get pod -o wide | grep 'upf' | awk '{"
                           "print $7}'")
            result = self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 0)
            if result[0] == result[1]:
                LOG.error('The pod-affinity rule doesn\'t worker.'
                          ' Please check your yaml file {}'.format(
                              upf_file_path))
                raise exceptions.MgmtDriverOtherError(
                    'The pod-affinity rule doesn\'t worker.'
                    ' Please check your yaml file {}'.format(
                        upf_file_path))
            else:
                LOG.debug('The pod has deployed on different worker node.')

    @log.log
    def heal_start(self, context, vnf_instance,
                   heal_vnf_request, grant,
                   grant_request, **kwargs):
        pass

    @log.log
    def heal_end(self, context, vnf_instance,
                 heal_vnf_request, grant,
                 grant_request, **kwargs):
        additional_param = vnf_instance.instantiated_vnf_info. \
            additional_params.get('free5gc', {})
        ssh_master_node_ip = additional_param.get('ssh_master_node_ip')
        master_node_username = additional_param.get('master_node_username')
        master_node_password = additional_param.get('master_node_password')
        if not additional_param.get('smf_config_file_path'):
            smf_config_file_path = '/go/src/free5gc/config/smfcfg.yaml'
        if not additional_param.get('upf_config_file_path'):
            upf_config_file_path = \
                '/go/src/free5gc/NFs/upf/build/config/upfcfg.yaml'
        commander = cmd_executer.RemoteCommandExecutor(
            user=master_node_username, password=master_node_password,
            host=ssh_master_node_ip,
            timeout=30)
        ssh_command = "kubectl get pod | grep upf | awk '{print $5}'"
        upf_pod_age_list = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT,
            'common', 3)
        if len(upf_pod_age_list) > 1:
            # get upf ip
            ssh_command = "kubectl get pod | grep smf | awk '{print $1}'"
            smf_pod_name = self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 3)[0].replace('\n', '')
            ssh_command = ("kubectl cp {smf_pod_name}:{smf_config_file_path}"
                           " /tmp/smfcfg.yaml -c myapp-container"
                           .format(smf_pod_name=smf_pod_name,
                                   smf_config_file_path=smf_config_file_path))
            self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 3)
            local_smf_path = '/tmp/smfcfg.yaml'
            self._send_and_receive_file(
                ssh_master_node_ip, master_node_username, master_node_password,
                '/tmp/smfcfg.yaml', local_smf_path, 'receive')
            with open(local_smf_path) as f:
                file_content = yaml.safe_load(f)
                upf1_pfcp_ip = file_content['configuration'][
                    'userplane_information']['up_nodes']['UPF']['node_id']
                upf2_pfcp_ip = file_content['configuration'][
                    'userplane_information']['up_nodes']['UPF2']['node_id']
            ssh_command = "kubectl get pod | grep upf | awk '{print $1}'"
            upf_pod_name_list = self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 3)
            for upf_pod in upf_pod_name_list:
                upf_example_file_path = \
                    '/go/src/free5gc/NFs/upf/build/config/upfcfg.yaml'
                ssh_command = "kubectl exec {} -c myapp-container --" \
                              " cat {}".format(upf_pod.replace('\n', ''),
                                               upf_example_file_path)
                results = self._execute_command(
                    commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                    'common', 3)
                for result in results:
                    if upf1_pfcp_ip in result:
                        flag = 'UPF2'
                        unhealed_upf_pod_name = upf_pod
                        break
                    if upf2_pfcp_ip in result:
                        flag = 'UPF'
                        unhealed_upf_pod_name = upf_pod
            upf_pod_name_list.remove(unhealed_upf_pod_name)
            upf_pod_name = upf_pod_name_list[0].replace('\n', '')
            upf_gtpu_list = []
            with open(local_smf_path) as f:
                file_content = yaml.safe_load(f)
                if flag == 'UPF':
                    upf_pfcp_ip = file_content['configuration'][
                        'userplane_information']['up_nodes']['UPF']['node_id']
                    upf_gtpu_interface_list = (
                        file_content['configuration']['userplane_information']
                        ['up_nodes']['UPF']['interfaces'])
                    for upf_gtpu_interface in upf_gtpu_interface_list:
                        upf_gtpu_list = (upf_gtpu_interface['endpoints'] +
                                         upf_gtpu_list)
                else:
                    upf_pfcp_ip = file_content['configuration'][
                        'userplane_information']['up_nodes']['UPF2']['node_id']
                    upf_gtpu_interface_list = (
                        file_content['configuration']['userplane_information']
                        ['up_nodes']['UPF2']['interfaces'])
                    for upf_gtpu_interface in upf_gtpu_interface_list:
                        upf_gtpu_list = (upf_gtpu_interface['endpoints'] +
                                         upf_gtpu_list)
            # modify upf config file
            upf_example_file_path = \
                '/go/src/free5gc/NFs/upf/build/config/upfcfg.yaml'
            ssh_command = ("kubectl cp {upf_pod_name}:{upf_example_file_path}"
                           " /tmp/upfcfg.yaml -c myapp-container".format(
                               upf_pod_name=upf_pod_name,
                               upf_example_file_path=upf_example_file_path))
            self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 3)
            local_upf_path = '/tmp/upfcfg.yaml'
            self._send_and_receive_file(
                ssh_master_node_ip, master_node_username, master_node_password,
                '/tmp/upfcfg.yaml', local_upf_path, 'receive')
            with open(local_upf_path, 'r') as f:
                upf_content = yaml.safe_load(f)
                upf_content['configuration']['pfcp'][0]['addr'] = upf_pfcp_ip
                for index in range(len(upf_gtpu_list)):
                    upf_content['configuration']['gtpu'][index]['addr'] = \
                        upf_gtpu_list[index]
            with open(local_upf_path, 'w') as nf:
                yaml.safe_dump(upf_content, nf, default_flow_style=False)
            self._send_and_receive_file(
                ssh_master_node_ip, master_node_username, master_node_password,
                '/tmp/upfcfg.yaml', local_upf_path, 'send')
            ssh_command = ("kubectl cp /tmp/upfcfg.yaml"
                           " {upf_pod_name}:{upf_config_file_path}"
                           " -c myapp-container"
                           .format(upf_pod_name=upf_pod_name,
                                   upf_config_file_path=upf_config_file_path))
            self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 3)

            # start upf process
            ssh_command = ('cat <<EOF | sudo tee "run_upf.sh" >/dev/null\n'
                           'kubectl exec'
                           ' {} -i -- sh'
                           '<< eof\n'
                           'ip link delete upfgtp\neof\n'
                           'kubectl exec'
                           ' {} -i -- sh'
                           '<< eof\n'
                           './NFs/upf/build/bin/free5gc-upfd -f {}\neof'
                           '\nEOF\n'
                           .format(upf_pod_name, upf_pod_name,
                                   upf_config_file_path))
            self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 3)
            ssh_command = "sudo chmod 777 run_upf.sh"
            self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 3)
            ssh_command = "nohup ./run_upf.sh > upf_heal.txt  2>&1 &"
            self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 3)

            # restart smf
            ssh_command = "kubectl get pod {} -o yaml | " \
                          "kubectl replace --force -f -".format(smf_pod_name)
            self._execute_command(
                commander, ssh_command, 120,
                'common', 0)
            time.sleep(120)
            ssh_command = "kubectl get pod | grep smf | awk '{print $1}'"
            smf_pod_name = self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 3)[0].replace('\n', '')
            ssh_command = "kubectl get pod {} | " \
                          "grep 'Running'".format(smf_pod_name)
            result = self._execute_command(
                commander, ssh_command, FREE5GC_CMD_TIMEOUT,
                'common', 0)
            if not result:
                LOG.error('SMF restart failed. Please check'
                          ' you k8s-cluster environment.')
                raise exceptions.MgmtDriverOtherError(
                    'SMF restart failed. Please check you'
                    ' k8s-cluster environment.')
            commander.close_session()
            os.remove(local_upf_path)
            os.remove(local_smf_path)
        else:
            self.instantiate_end(context, vnf_instance,
                                 vnf_instance.instantiated_vnf_info, grant,
                                 grant_request, **kwargs)

    @log.log
    def change_external_connectivity_start(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass

    @log.log
    def change_external_connectivity_end(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass



# Copyright 2016 Brocade Communications System, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock


from tacker.common import exceptions
from tacker import context
from tacker.db.migration import purge_tables
from tacker.tests.unit.db import base as db_base


class FakeConfig(mock.Mock):
    pass


class TestDbPurgeDelete(db_base.SqlTestCase):
    def setUp(self):
        super(TestDbPurgeDelete, self).setUp()
        self.addCleanup(mock.patch.stopall)
        self.context = context.get_admin_context()
        self._mock_config()
        mock.patch('sqlalchemy.Table').start()
        mock.patch('tacker.db.migration.purge_tables._purge_resource_tables'
                   ).start()
        mock.patch('tacker.db.migration.purge_tables._purge_events_table',
                   ).start()
        mock.patch('tacker.db.migration.purge_tables.'
                   '_generate_associated_tables_map').start()
        mock.patch('tacker.db.migration.purge_tables.get_engine').start()

    def _mock_config(self):
        self.config = mock.Mock(wraps=FakeConfig())
        fake_config = mock.Mock()
        fake_config.return_value = self.config
        self._mock(
            'alembic.config.__init__', fake_config)

    def test_age_not_integer_input(self):
        self.assertRaises(exceptions.InvalidInput, purge_tables.purge_deleted,
                          self.config, 'invalid', 'abc')

    def test_age_negative_integer_input(self):
        self.assertRaises(exceptions.InvalidInput, purge_tables.purge_deleted,
                          self.config, 'invalid', '-90')

    def test_invalid_granularity_input(self):
        self.assertRaises(exceptions.InvalidInput, purge_tables.purge_deleted,
                          self.config, 'vnf', '90', 'decade')

    def test_purge_delete_call_vnf(self):
        purge_tables.purge_deleted(self.config, 'vnf', '90', 'days')
        purge_tables._purge_resource_tables.assert_called_once_with(
            mock.ANY, mock.ANY, mock.ANY, mock.ANY, mock.ANY)

    def test_purge_delete_call_vnfd(self):
        purge_tables.purge_deleted(self.config, 'vnfd', '90', 'days')
        purge_tables._purge_resource_tables.assert_called_once_with(
            mock.ANY, mock.ANY, mock.ANY, mock.ANY, mock.ANY)

    def test_purge_delete_call_vim(self):
        purge_tables.purge_deleted(self.config, 'vims', '90', 'days')
        purge_tables._purge_resource_tables.assert_called_once_with(
            mock.ANY, mock.ANY, mock.ANY, mock.ANY, mock.ANY)

    def test_purge_delete_call_events(self):
        purge_tables.purge_deleted(self.config, 'events', '90', 'days')
        purge_tables._purge_events_table.assert_called_once_with(
            mock.ANY, mock.ANY, mock.ANY)



# Copyright (c) 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_log import log as logging
from oslo_serialization import jsonutils
import routes
import webob
import webtest

from tacker.api import extensions
from tacker.common import config
from tacker.plugins.common import constants
from tacker.tests import base
from tacker.tests.unit import extension_stubs as ext_stubs
import tacker.tests.unit.extensions
from tacker.tests.unit import testlib_api
from tacker import wsgi


LOG = logging.getLogger(__name__)
extensions_path = ':'.join(tacker.tests.unit.extensions.__path__)


class ExtensionsTestApp(wsgi.Router):

    def __init__(self, options={}):
        mapper = routes.Mapper()
        controller = ext_stubs.StubBaseAppController()
        mapper.resource("dummy_resource", "/dummy_resources",
                        controller=controller)
        super(ExtensionsTestApp, self).__init__(mapper)


class PluginInterfaceTest(base.BaseTestCase):
    def test_issubclass_hook(self):
        class A(object):
            def f(self):
                pass

        class B(extensions.PluginInterface):
            @abc.abstractmethod
            def f(self):
                pass

        self.assertTrue(issubclass(A, B))

    def test_issubclass_hook_class_without_abstract_methods(self):
        class A(object):
            def f(self):
                pass

        class B(extensions.PluginInterface):
            def f(self):
                pass

        self.assertFalse(issubclass(A, B))

    def test_issubclass_hook_not_all_methods_implemented(self):
        class A(object):
            def f(self):
                pass

        class B(extensions.PluginInterface):
            @abc.abstractmethod
            def f(self):
                pass

            @abc.abstractmethod
            def g(self):
                pass

        self.assertFalse(issubclass(A, B))


class ResourceExtensionTest(base.BaseTestCase):

    class ResourceExtensionController(wsgi.Controller):

        def index(self, request):
            return "resource index"

        def show(self, request, id):
            return {'data': {'id': id}}

        def notimplemented_function(self, request, id):
            return webob.exc.HTTPNotImplemented()

        def custom_member_action(self, request, id):
            return {'member_action': 'value'}

        def custom_collection_action(self, request, **kwargs):
            return {'collection': 'value'}

    class DummySvcPlugin(wsgi.Controller):
        def get_plugin_type(self):
            return constants.DUMMY

        def index(self, request, **kwargs):
            return "resource index"

        def custom_member_action(self, request, **kwargs):
            return {'member_action': 'value'}

        def collection_action(self, request, **kwargs):
            return {'collection': 'value'}

        def show(self, request, id):
            return {'data': {'id': id}}

    def test_exceptions_notimplemented(self):
        controller = self.ResourceExtensionController()
        member = {'notimplemented_function': "GET"}
        res_ext = extensions.ResourceExtension('tweedles', controller,
                                               member_actions=member)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        # Ideally we would check for a 501 code here but webtest doesn't take
        # anything that is below 200 or above 400 so we can't actually check
        # it.  It throws webtest.AppError instead.
        try:
            test_app.get("/tweedles/some_id/notimplemented_function")
            # Shouldn't be reached
            self.assertTrue(False)
        except webtest.AppError as e:
            self.assertIn('501', str(e))

    def test_resource_can_be_added_as_extension(self):
        res_ext = extensions.ResourceExtension(
            'tweedles', self.ResourceExtensionController())
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))
        index_response = test_app.get("/tweedles")
        self.assertEqual(200, index_response.status_int)
        self.assertEqual(b"resource index", index_response.body)

        show_response = test_app.get("/tweedles/25266")
        self.assertEqual({'data': {'id': "25266"}}, show_response.json)

    def test_resource_gets_prefix_of_plugin(self):
        class DummySvcPlugin(wsgi.Controller):
            def index(self, request):
                return ""

            def get_plugin_type(self):
                return constants.DUMMY

        res_ext = extensions.ResourceExtension(
            'tweedles', DummySvcPlugin(), path_prefix="/dummy_svc")
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))
        index_response = test_app.get("/dummy_svc/tweedles")
        self.assertEqual(200, index_response.status_int)

    def test_resource_extension_with_custom_member_action(self):
        controller = self.ResourceExtensionController()
        member = {'custom_member_action': "GET"}
        res_ext = extensions.ResourceExtension('tweedles', controller,
                                               member_actions=member)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        response = test_app.get("/tweedles/some_id/custom_member_action")
        self.assertEqual(200, response.status_int)
        self.assertEqual("value",
                         jsonutils.loads(response.body)['member_action'])

    def test_resource_ext_with_custom_member_action_gets_plugin_prefix(self):
        controller = self.DummySvcPlugin()
        member = {'custom_member_action': "GET"}
        collections = {'collection_action': "GET"}
        res_ext = extensions.ResourceExtension('tweedles', controller,
                                               path_prefix="/dummy_svc",
                                               member_actions=member,
                                               collection_actions=collections)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        response = test_app.get("/dummy_svc/tweedles/1/custom_member_action")
        self.assertEqual(200, response.status_int)
        self.assertEqual("value",
                         jsonutils.loads(response.body)['member_action'])

        response = test_app.get("/dummy_svc/tweedles/collection_action")
        self.assertEqual(200, response.status_int)
        self.assertEqual("value",
                         jsonutils.loads(response.body)['collection'])

    def test_plugin_prefix_with_parent_resource(self):
        controller = self.DummySvcPlugin()
        parent = dict(member_name="tenant",
                      collection_name="tenants")
        member = {'custom_member_action': "GET"}
        collections = {'collection_action': "GET"}
        res_ext = extensions.ResourceExtension('tweedles', controller, parent,
                                               path_prefix="/dummy_svc",
                                               member_actions=member,
                                               collection_actions=collections)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        index_response = test_app.get("/dummy_svc/tenants/1/tweedles")
        self.assertEqual(200, index_response.status_int)

        response = test_app.get("/dummy_svc/tenants/1/"
                                "tweedles/1/custom_member_action")
        self.assertEqual(200, response.status_int)
        self.assertEqual(jsonutils.loads(response.body)['member_action'],
                         "value")

        response = test_app.get("/dummy_svc/tenants/2/"
                                "tweedles/collection_action")
        self.assertEqual(200, response.status_int)
        self.assertEqual("value",
                         jsonutils.loads(response.body)['collection'])

    def test_resource_extension_for_get_custom_collection_action(self):
        controller = self.ResourceExtensionController()
        collections = {'custom_collection_action': "GET"}
        res_ext = extensions.ResourceExtension('tweedles', controller,
                                               collection_actions=collections)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        response = test_app.get("/tweedles/custom_collection_action")
        self.assertEqual(200, response.status_int)
        LOG.debug(jsonutils.loads(response.body))
        self.assertEqual("value", jsonutils.loads(response.body)['collection'])

    def test_resource_extension_for_put_custom_collection_action(self):
        controller = self.ResourceExtensionController()
        collections = {'custom_collection_action': "PUT"}
        res_ext = extensions.ResourceExtension('tweedles', controller,
                                               collection_actions=collections)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        response = test_app.put("/tweedles/custom_collection_action")

        self.assertEqual(200, response.status_int)
        self.assertEqual('value', jsonutils.loads(response.body)['collection'])

    def test_resource_extension_for_post_custom_collection_action(self):
        controller = self.ResourceExtensionController()
        collections = {'custom_collection_action': "POST"}
        res_ext = extensions.ResourceExtension('tweedles', controller,
                                               collection_actions=collections)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        response = test_app.post("/tweedles/custom_collection_action")

        self.assertEqual(200, response.status_int)
        self.assertEqual('value', jsonutils.loads(response.body)['collection'])

    def test_resource_extension_for_delete_custom_collection_action(self):
        controller = self.ResourceExtensionController()
        collections = {'custom_collection_action': "DELETE"}
        res_ext = extensions.ResourceExtension('tweedles', controller,
                                               collection_actions=collections)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        response = test_app.delete("/tweedles/custom_collection_action")

        self.assertEqual(200, response.status_int)
        self.assertEqual('value', jsonutils.loads(response.body)['collection'])

    def test_resource_ext_for_formatted_req_on_custom_collection_action(self):
        controller = self.ResourceExtensionController()
        collections = {'custom_collection_action': "GET"}
        res_ext = extensions.ResourceExtension('tweedles', controller,
                                               collection_actions=collections)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        response = test_app.get("/tweedles/custom_collection_action.json")

        self.assertEqual(200, response.status_int)
        self.assertEqual('value', jsonutils.loads(response.body)['collection'])

    def test_resource_ext_for_nested_resource_custom_collection_action(self):
        controller = self.ResourceExtensionController()
        collections = {'custom_collection_action': "GET"}
        parent = dict(collection_name='beetles', member_name='beetle')
        res_ext = extensions.ResourceExtension('tweedles', controller,
                                               collection_actions=collections,
                                               parent=parent)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        response = test_app.get("/beetles/beetle_id"
                                "/tweedles/custom_collection_action")

        self.assertEqual(200, response.status_int)
        self.assertEqual('value', jsonutils.loads(response.body)['collection'])

    def test_resource_extension_with_custom_member_action_and_attr_map(self):
        controller = self.ResourceExtensionController()
        member = {'custom_member_action': "GET"}
        params = {
            'tweedles': {
                'id': {'allow_post': False, 'allow_put': False,
                       'validate': {'type:uuid': None},
                       'is_visible': True},
                'name': {'allow_post': True, 'allow_put': True,
                         'validate': {'type:string': None},
                         'default': '', 'is_visible': True},
            }
        }
        res_ext = extensions.ResourceExtension('tweedles', controller,
                                               member_actions=member,
                                               attr_map=params)
        test_app = _setup_extensions_test_app(SimpleExtensionManager(res_ext))

        response = test_app.get("/tweedles/some_id/custom_member_action")
        self.assertEqual(200, response.status_int)
        self.assertEqual('value',
                         jsonutils.loads(response.body)['member_action'])

    def test_returns_404_for_non_existent_extension(self):
        test_app = _setup_extensions_test_app(SimpleExtensionManager(None))

        response = test_app.get("/non_extistant_extension", status='*')

        self.assertEqual(404, response.status_int)


class ActionExtensionTest(base.BaseTestCase):

    def setUp(self):
        super(ActionExtensionTest, self).setUp()
        self.skipTest("Not ready yet")
        self.extension_app = _setup_extensions_test_app()

    def test_extended_action_for_adding_extra_data(self):
        action_name = 'FOXNSOX:add_tweedle'
        action_params = dict(name='Beetle')
        req_body = jsonutils.dump_as_bytes({action_name: action_params})
        response = self.extension_app.post('/dummy_resources/1/action',
                                           req_body,
                                           content_type='application/json')
        self.assertEqual("Tweedle Beetle Added.", response.body)

    def test_extended_action_for_deleting_extra_data(self):
        action_name = 'FOXNSOX:delete_tweedle'
        action_params = dict(name='Bailey')
        req_body = jsonutils.dump_as_bytes({action_name: action_params})
        response = self.extension_app.post("/dummy_resources/1/action",
                                           req_body,
                                           content_type='application/json')
        self.assertEqual("Tweedle Bailey Deleted.", response.body)

    def test_returns_404_for_non_existent_action(self):
        non_existent_action = 'blah_action'
        action_params = dict(name="test")
        req_body = jsonutils.dump_as_bytes(
            {non_existent_action: action_params})

        response = self.extension_app.post("/dummy_resources/1/action",
                                           req_body,
                                           content_type='application/json',
                                           status='*')

        self.assertEqual(404, response.status_int)

    def test_returns_404_for_non_existent_resource(self):
        action_name = 'add_tweedle'
        action_params = dict(name='Beetle')
        req_body = jsonutils.dump_as_bytes({action_name: action_params})

        response = self.extension_app.post("/asdf/1/action", req_body,
                                           content_type='application/json',
                                           status='*')
        self.assertEqual(404, response.status_int)


class RequestExtensionTest(base.BaseTestCase):

    def test_headers_can_be_extended(self):
        def extend_headers(req, res):
            assert req.headers['X-NEW-REQUEST-HEADER'] == "sox"
            res.headers['X-NEW-RESPONSE-HEADER'] = "response_header_data"
            return res

        app = self._setup_app_with_request_handler(extend_headers, 'GET')
        response = app.get("/dummy_resources/1",
                           headers={'X-NEW-REQUEST-HEADER': "sox"})

        self.assertEqual("response_header_data",
                         response.headers['X-NEW-RESPONSE-HEADER'])

    def test_extend_get_resource_response(self):
        def extend_response_data(req, res):
            data = jsonutils.loads(res.body)
            data['FOXNSOX:extended_key'] = req.GET.get('extended_key')
            res.body = jsonutils.dump_as_bytes(data)
            return res

        app = self._setup_app_with_request_handler(extend_response_data, 'GET')
        response = app.get("/dummy_resources/1?extended_key=extended_data")

        self.assertEqual(200, response.status_int)
        response_data = jsonutils.loads(response.body)
        self.assertEqual('extended_data',
                         response_data['FOXNSOX:extended_key'])
        self.assertEqual('knox', response_data['fort'])

    def test_get_resources(self):
        self.skipTest("Not ready yet")
        app = _setup_extensions_test_app()

        response = app.get("/dummy_resources/1?chewing=newblue")

        response_data = jsonutils.loads(response.body)
        self.assertEqual('newblue', response_data['FOXNSOX:googoose'])
        self.assertEqual("Pig Bands!", response_data['FOXNSOX:big_bands'])

    def test_edit_previously_uneditable_field(self):

        def _update_handler(req, res):
            data = jsonutils.loads(res.body)
            data['uneditable'] = req.params['uneditable']
            res.body = jsonutils.dump_as_bytes(data)
            return res

        base_app = webtest.TestApp(setup_base_app(self))
        response = base_app.put("/dummy_resources/1",
                                {'uneditable': "new_value"})
        self.assertEqual("original_value", response.json['uneditable'])

        ext_app = self._setup_app_with_request_handler(_update_handler,
                                                       'PUT')
        ext_response = ext_app.put("/dummy_resources/1",
                                   {'uneditable': "new_value"})
        self.assertEqual("new_value", ext_response.json['uneditable'])

    def _setup_app_with_request_handler(self, handler, verb):
        req_ext = extensions.RequestExtension(verb,
                                              '/dummy_resources/:(id)',
                                              handler)
        manager = SimpleExtensionManager(None, None, req_ext)
        return _setup_extensions_test_app(manager)


class ExtensionManagerTest(base.BaseTestCase):

    def test_invalid_extensions_are_not_registered(self):

        class InvalidExtension(object):
            """Invalid extension.

            This Extension doesn't implement extension methods :
            get_name, get_description, get_namespace and get_updated
            """
            def get_alias(self):
                return "invalid_extension"

        ext_mgr = extensions.ExtensionManager('')
        ext_mgr.add_extension(InvalidExtension())
        ext_mgr.add_extension(ext_stubs.StubExtension("valid_extension"))

        self.assertIn('valid_extension', ext_mgr.extensions)
        self.assertNotIn('invalid_extension', ext_mgr.extensions)


class ExtensionControllerTest(testlib_api.WebTestCase):

    def setUp(self):
        super(ExtensionControllerTest, self).setUp()
        self.skipTest("Not ready yet")
        self.test_app = _setup_extensions_test_app()

    def test_index_gets_all_registerd_extensions(self):
        response = self.test_app.get("/extensions." + self.fmt)
        res_body = self.deserialize(response)
        foxnsox = res_body["extensions"][0]

        self.assertEqual("FOXNSOX", foxnsox["alias"])
        self.assertEqual("http://www.fox.in.socks/api/ext/pie/v1.0",
                         foxnsox["namespace"])

    def test_extension_can_be_accessed_by_alias(self):
        response = self.test_app.get("/extensions/FOXNSOX." + self.fmt)
        foxnsox_extension = self.deserialize(response)
        foxnsox_extension = foxnsox_extension['extension']
        self.assertEqual("FOXNSOX", foxnsox_extension["alias"])
        self.assertEqual("http://www.fox.in.socks/api/ext/pie/v1.0",
                         foxnsox_extension["namespace"])

    def test_show_returns_not_found_for_non_existent_extension(self):
        response = self.test_app.get("/extensions/non_existent" + self.fmt,
                                     status="*")

        self.assertEqual(404, response.status_int)


def app_factory(global_conf, **local_conf):
    conf = global_conf.copy()
    conf.update(local_conf)
    return ExtensionsTestApp(conf)


def setup_base_app(test):
    base.BaseTestCase.config_parse()
    app = config.load_paste_app('extensions_test_app')
    return app


def setup_extensions_middleware(extension_manager=None):
    extension_manager = (extension_manager or
                         extensions.ExtensionManager(extensions_path))
    base.BaseTestCase.config_parse()
    app = config.load_paste_app('extensions_test_app')
    return extensions.ExtensionMiddleware(app, ext_mgr=extension_manager)


def _setup_extensions_test_app(extension_manager=None):
    return webtest.TestApp(setup_extensions_middleware(extension_manager))


class SimpleExtensionManager(object):

    def __init__(self, resource_ext=None, action_ext=None, request_ext=None):
        self.resource_ext = resource_ext
        self.action_ext = action_ext
        self.request_ext = request_ext

    def get_resources(self):
        resource_exts = []
        if self.resource_ext:
            resource_exts.append(self.resource_ext)
        return resource_exts

    def get_actions(self):
        action_exts = []
        if self.action_ext:
            action_exts.append(self.action_ext)
        return action_exts

    def get_request_extensions(self):
        request_extensions = []
        if self.request_ext:
            request_extensions.append(self.request_ext)
        return request_extensions



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import codecs
import os
from unittest import mock
import yaml

from oslo_serialization import jsonutils

from tacker import context
from tacker.db.common_services import common_services_db_plugin
from tacker.extensions import vnfm
from tacker.objects import fields
from tacker.tests.unit import base
from tacker.tests.unit.db import utils
from tacker.vnfm.infra_drivers.openstack import openstack


vnf_dict = {
    'attributes': {
        'heat_template': {
            'outputs': {
                'mgmt_ip-VDU1': {
                    'value': {
                        'get_attr': [
                            'CP1', 'fixed_ips', 0, 'ip_address']
                    }
                }
            },
            'description': 'Demo example\n',
            'parameters': {},
            'resources': {
                'VDU1': {
                    'type': 'OS::Nova::Server',
                    'properties': {
                        'user_data_format': 'SOFTWARE_CONFIG',
                        'availability_zone': 'nova',
                        'image': 'cirros-0.5.2-x86_64-disk',
                        'config_drive': False,
                        'flavor': {'get_resource': 'VDU1_flavor'},
                        'networks': [{'port': {'get_resource': 'CP1'}}]
                    }
                },
                'CP1': {
                    'type': 'OS::Neutron::Port',
                    'properties': {
                        'port_security_enabled': False,
                        'network': 'net_mgmt'
                    }
                },
                'VDU1_flavor': {
                    'type': 'OS::Nova::Flavor',
                    'properties': {'vcpus': 1, 'disk': 1, 'ram': 512}
                }
            }
        }
    },
    'status': 'ACTIVE',
    'vnfd_id': '576acf48-b9df-491d-a57c-342de660ec78',
    'tenant_id': '13d2ca8de70d48b2a2e0dbac2c327c0b',
    'vim_id': '3f41faa7-5630-47d2-9d4a-1216953c8887',
    'instance_id': 'd1121d3c-368b-4ac2-b39d-835aa3e4ccd8',
    'placement_attr': {'vim_name': 'openstack-vim'},
    'id': 'a27fc58e-66ae-4031-bba4-efede318c60b',
    'name': 'vnf_create_1'
}


class FakeHeatClient(mock.Mock):

    class Stack(mock.Mock):
        stack_status = 'CREATE_COMPLETE'
        outputs = [{'output_value': '192.168.120.31', 'description':
            'management ip address', 'output_key': 'mgmt_ip-vdu1'}]

    def create(self, *args, **kwargs):
        return {'stack': {'id': '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}}

    def get(self, id):
        return self.Stack()


def _get_template(name):
    filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "data/", name)
    with codecs.open(filename, encoding='utf-8', errors='strict') as f:
        return f.read()


class TestOpenStack(base.TestCase):
    hot_template = _get_template('hot_openwrt.yaml')
    hot_param_template = _get_template('hot_openwrt_params.yaml')
    hot_ipparam_template = _get_template('hot_openwrt_ipparams.yaml')
    tosca_vnfd_openwrt = _get_template('test_tosca_openwrt.yaml')
    tosca_vnfd_openwrt_param = _get_template('test_tosca_openwrt_param.yaml')
    config_data = _get_template('config_data.yaml')

    def setUp(self):
        super(TestOpenStack, self).setUp()
        self.context = context.get_admin_context()
        self.infra_driver = openstack.OpenStack()
        self._mock_heat_client()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin = \
            common_services_db_plugin.CommonServicesPluginDb()
        self.addCleanup(mock.patch.stopall)
        yaml.SafeLoader.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            lambda loader, node: dict(loader.construct_pairs(node)))

    def _mock_heat_client(self):
        self.heat_client = mock.Mock(wraps=FakeHeatClient())
        fake_heat_client = mock.Mock()
        fake_heat_client.return_value = self.heat_client
        self._mock(
            'tacker.vnfm.infra_drivers.openstack.heat_client.HeatClient',
            fake_heat_client)

    def _mock(self, target, new=mock.DEFAULT):
        patcher = mock.patch(target, new)
        return patcher.start()

    def _get_vnfd(self, template):
        return {'vnfd': {'attributes': {'vnfd': template}}}

    def _get_expected_vnfd(self, template):
        return {'attributes': {'vnfd': template},
                'description': 'OpenWRT with services',
                'mgmt_driver': 'openwrt', 'name': 'OpenWRT',
                'service_types': [{'service_type': 'vnfd',
                'id': '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}],
                'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                'id': 'fb048660-dc1b-4f0f-bd89-b023666650ec'}

    def _get_expected_vnf_update_obj(self):
        return {'status': 'PENDING_CREATE', 'instance_id': None, 'name':
             'test_openwrt', 'tenant_id':
        'ad7ebc56538745a08ef7c5e97f8bd437', 'vnfd_id':
        'eb094833-995e-49f0-a047-dfb56aaf7c4e', 'vnfd': {
            'service_types': [{'service_type': 'vnfd', 'id':
            '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}], 'description':
            'OpenWRT with services', 'tenant_id':
            'ad7ebc56538745a08ef7c5e97f8bd437', 'mgmt_driver': 'openwrt',
            'attributes': {'vnfd': self.tosca_vnfd_openwrt},
            'id': 'fb048660-dc1b-4f0f-bd89-b023666650ec', 'name':
            'openwrt_services'}, 'mgmt_ip_address': None,
            'service_context': [],
            'attributes': {'config': utils.update_config_data},
            'id': 'eb84260e-5ff7-4332-b032-50a14d6c1123', 'description':
                'OpenWRT with services'}

    def _get_expected_vnf_update_new_param_obj(self):
        return {'status': 'PENDING_CREATE', 'instance_id': None, 'name':
            'test_openwrt', 'tenant_id':
        'ad7ebc56538745a08ef7c5e97f8bd437', 'vnfd_id':
        'eb094833-995e-49f0-a047-dfb56aaf7c4e', 'vnfd': {
            'service_types': [{'service_type': 'vnfd', 'id':
            '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}], 'description':
            'OpenWRT with services', 'tenant_id':
            'ad7ebc56538745a08ef7c5e97f8bd437', 'mgmt_driver': 'openwrt',
            'attributes': {'vnfd': self.tosca_vnfd_openwrt_param},
            'id': 'fb048660-dc1b-4f0f-bd89-b023666650ec', 'name':
            'openwrt_services'}, 'mgmt_url': None, 'service_context': [],
            'attributes': {'heat_template': utils.hot_data,
                           'param_values': utils.update_new_param_data},
            'id': 'eb84260e-5ff7-4332-b032-50a14d6c1123', 'description':
                'OpenWRT with services'}

    def _get_expected_active_vnf(self):
        return {'status': 'ACTIVE',
                'instance_id': None,
                'name': 'test_openwrt',
                'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                'vnfd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
                'vnfd': {
                    'service_types': [{
                        'service_type': 'vnfd',
                        'id': '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}],
                    'description': 'OpenWRT with services',
                    'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                    'mgmt_driver': 'openwrt',
                    'infra_driver': 'heat',
                    'attributes': {'vnfd': self.tosca_vnfd_openwrt},
                    'id': 'fb048660-dc1b-4f0f-bd89-b023666650ec',
                    'name': 'openwrt_services'},
                'mgmt_ip_address': '{"vdu1": "192.168.120.31"}',
                'service_context': [],
                'id': 'eb84260e-5ff7-4332-b032-50a14d6c1123',
                'description': 'OpenWRT with services'}

    def test_delete(self):
        vnf_id = '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'
        self.infra_driver.delete(plugin=None, context=self.context,
                                vnf_id=vnf_id,
                                auth_attr=utils.get_vim_auth_obj())
        self.heat_client.delete.assert_called_once_with(vnf_id)

    def test_update(self):
        vnf_obj = utils.get_dummy_vnf_config_attr()
        vnf_config_obj = utils.get_dummy_vnf_update_config()
        expected_vnf_update = self._get_expected_vnf_update_obj()
        vnf_id = '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'
        self.infra_driver.update(plugin=None, context=self.context,
                                 vnf_id=vnf_id, vnf_dict=vnf_obj,
                                 vnf=vnf_config_obj,
                                 auth_attr=utils.get_vim_auth_obj())
        expected_vnf_update['attributes']['config'] = yaml.safe_load(
            expected_vnf_update['attributes']['config'])
        vnf_obj['attributes']['config'] = yaml.safe_load(vnf_obj['attributes'][
            'config'])
        self.assertEqual(expected_vnf_update, vnf_obj)

    @mock.patch(
        'tacker.vnfm.infra_drivers.openstack.vdu.Vdu')
    def test_heal_vdu(self, mock_vdu):
        self.infra_driver.heal_vdu(None, self.context, vnf_dict,
                                   mock.ANY)
        mock_vdu.assert_called_once_with(self.context, vnf_dict,
                                         mock.ANY)

    @mock.patch(
        'tacker.vnfm.infra_drivers.openstack.vdu.Vdu')
    @mock.patch('tacker.vnfm.infra_drivers.openstack.openstack.LOG')
    def test_heal_vdu_failed(self, mock_log, mock_vdu):
        mock_vdu.side_effect = Exception
        self.assertRaises(vnfm.VNFHealFailed, self.infra_driver.heal_vdu,
                          None, self.context, vnf_dict,
                          mock.ANY)
        mock_log.error.assert_called_with(
            "VNF '%s' failed to heal", vnf_dict['id'])

    def test_update_new_param(self):
        vnf_obj = utils.get_dummy_vnf_param_attr()
        vnf_param_obj = utils.get_dummy_vnf_update_new_param()
        expected_vnf_update = self._get_expected_vnf_update_new_param_obj()
        vnf_id = '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'
        self.infra_driver.update(plugin=None, context=self.context,
                                 vnf_id=vnf_id, vnf_dict=vnf_obj,
                                 vnf=vnf_param_obj,
                                 auth_attr=utils.get_vim_auth_obj())
        expected_vnf_update['attributes']['param_values'] = yaml.safe_load(
            expected_vnf_update['attributes']['param_values'])
        vnf_obj['attributes']['param_values'] = yaml.safe_load(
            vnf_obj['attributes']['param_values'])
        self.assertEqual(expected_vnf_update, vnf_obj)

    @mock.patch('tacker.vnfm.infra_drivers.openstack.openstack.LOG')
    def test_update_invalid_param(self, mock_log):
        vnf_obj = utils.get_dummy_vnf_param_attr()
        vnf_param_obj = utils.get_dummy_vnf_update_invalid_param()
        vnf_id = '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'
        self.assertRaises(vnfm.VNFUpdateInvalidInput,
                          self.infra_driver.update,
                          plugin=None, context=self.context,
                          vnf_id=vnf_id, vnf_dict=vnf_obj,
                          vnf=vnf_param_obj,
                          auth_attr=utils.get_vim_auth_obj())
        log_msg = "at vnf_id {} because all parameters "\
                  "match the existing one.".format(vnf_id)
        mock_log.warning.assert_called_with(log_msg)

    @mock.patch('tacker.vnfm.infra_drivers.openstack.openstack.LOG')
    def test_update_empty_param(self, mock_log):
        vnf_obj = utils.get_dummy_vnf_param_attr()
        vnf_param_obj = utils.get_dummy_vnf_update_empty_param()
        vnf_id = '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'
        self.assertRaises(vnfm.VNFUpdateInvalidInput,
                          self.infra_driver.update,
                          plugin=None, context=self.context,
                          vnf_id=vnf_id, vnf_dict=vnf_obj,
                          vnf=vnf_param_obj,
                          auth_attr=utils.get_vim_auth_obj())
        log_msg = "at vnf_id {} because the target "\
                  "yaml is empty.".format(vnf_id)
        mock_log.warning.assert_called_with(log_msg)

    def _get_expected_fields_tosca(self, template):
        return {'stack_name':
                'test_openwrt_eb84260e-5ff7-4332-b032-50a14d6c1123',
                'template': _get_template(template),
                'timeout_mins': 10}

    def _get_expected_tosca_vnf(self,
                                tosca_tpl_name,
                                hot_tpl_name,
                                param_values='',
                                is_monitor=True,
                                multi_vdus=False):
        tosca_tpl = _get_template(tosca_tpl_name)
        exp_tmpl = self._get_expected_vnfd(tosca_tpl)
        tosca_hw_dict = yaml.safe_load(_get_template(hot_tpl_name))
        dvc = {
            'vnfd': exp_tmpl,
            'description': 'OpenWRT with services',
            'attributes': {
                'heat_template': tosca_hw_dict,
                'param_values': param_values
            },
            'id': 'eb84260e-5ff7-4332-b032-50a14d6c1123',
            'instance_id': None,
            'mgmt_ip_address': None,
            'name': 'test_openwrt',
            'service_context': [],
            'status': 'PENDING_CREATE',
            'vnfd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
            'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437'
        }
        # Add monitoring attributes for those yaml, which are having it
        if is_monitor:
            if multi_vdus:
                dvc['attributes'].update(
                    {'monitoring_policy': bytes(
                     '{"vdus": {"VDU1": {"ping": '
                     '{"name": "ping", "actions": '
                     '{"failure": "respawn"}, '
                     '"parameters": {"count": 3, '
                     '"interval": 10}, '
                     '"monitoring_params": '
                     '{"count": 3, "interval": 10}}}, '
                     '"VDU2": {"ping": {"name": "ping", '
                     '"actions": {"failure": "respawn"}, '
                     '"parameters": {"count": 3, '
                     '"interval": 10}, '
                     '"monitoring_params": {"count": 3, '
                     '"interval": 10}}}}}', 'utf-8')})
            else:
                dvc['attributes'].update(
                    {'monitoring_policy': bytes(
                     '{"vdus": {"VDU1": {"ping": '
                     '{"name": "ping", "actions": '
                     '{"failure": "respawn"}, '
                     '"parameters": {"count": 3, '
                     '"interval": 10}, '
                     '"monitoring_params": '
                     '{"count": 3, '
                     '"interval": 10}}}}}', 'utf-8')})

        return dvc

    def _get_dummy_tosca_vnf(self, template, input_params=''):

        tosca_template = _get_template(template)
        vnf = utils.get_dummy_vnf()
        dtemplate = self._get_expected_vnfd(tosca_template)

        vnf['vnfd'] = dtemplate
        vnf['attributes'] = {}
        vnf['attributes']['param_values'] = input_params
        return vnf

    def _test_assert_equal_for_tosca_templates(self,
                                               tosca_tpl_name,
                                               hot_tpl_name,
                                               input_params='',
                                               files=None,
                                               is_monitor=True,
                                               multi_vdus=False):
        vnf = self._get_dummy_tosca_vnf(tosca_tpl_name, input_params)
        expected_result = '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'
        expected_fields = self._get_expected_fields_tosca(hot_tpl_name)
        expected_vnf = self._get_expected_tosca_vnf(tosca_tpl_name,
                                                    hot_tpl_name,
                                                    input_params,
                                                    is_monitor,
                                                    multi_vdus)
        vnf['before_error_point'] = fields.ErrorPoint.PRE_VIM_CONTROL
        result = self.infra_driver.create(plugin=None, context=self.context,
                                         vnf=vnf,
                                         auth_attr=utils.get_vim_auth_obj())
        del vnf['before_error_point']
        actual_fields = self.heat_client.create.call_args[0][0]
        actual_fields["template"] = yaml.safe_load(actual_fields["template"])
        expected_fields["template"] = \
            yaml.safe_load(expected_fields["template"])

        if files:
            for k, v in actual_fields["files"].items():
                actual_fields["files"][k] = yaml.safe_load(v)

            expected_fields["files"] = {}
            for k, v in files.items():
                expected_fields["files"][k] = yaml.safe_load(_get_template(v))

        self.assertEqual(expected_fields, actual_fields)
        vnf["attributes"]["heat_template"] = yaml.safe_load(
            vnf["attributes"]["heat_template"])
        self.heat_client.create.assert_called_once_with(expected_fields)
        self.assertEqual(expected_result, result)

        if files:
            expected_fields["files"] = {}
            for k, v in files.items():
                expected_vnf["attributes"][k] = yaml.safe_load(
                    _get_template(v))
                vnf["attributes"][k] = yaml.safe_load(
                    vnf["attributes"][k])
            expected_vnf["attributes"]['scaling_group_names'] = {
                'SP1': 'SP1_group'}
            vnf["attributes"]['scaling_group_names'] = jsonutils.loads(
                vnf["attributes"]['scaling_group_names']
            )
        self.assertEqual(expected_vnf, vnf)

    def test_create_tosca(self):
        # self.skipTest("Not ready yet")
        self._test_assert_equal_for_tosca_templates('test_tosca_openwrt.yaml',
            'hot_tosca_openwrt.yaml')

    def test_create_tosca_with_userdata(self):
        self._test_assert_equal_for_tosca_templates(
            'test_tosca_openwrt_userdata.yaml',
            'hot_tosca_openwrt_userdata.yaml')

    def test_create_tosca_with_new_flavor(self):
        self._test_assert_equal_for_tosca_templates('test_tosca_flavor.yaml',
            'hot_flavor.yaml')

    def test_create_tosca_with_new_flavor_with_defaults(self):
        self._test_assert_equal_for_tosca_templates(
            'test_tosca_flavor_defaults.yaml',
            'hot_flavor_defaults.yaml')

    def test_create_tosca_with_placement_policy_anti_affinity(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_placement_policy_anti_affinity.yaml',
            'hot_placement_policy_anti_affinity.yaml', is_monitor=False)

    def test_create_tosca_with_placement_policy_affinity(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_placement_policy_affinity.yaml',
            'hot_placement_policy_affinity.yaml', is_monitor=False)

    def test_create_tosca_with_placement_policy_soft_anti_affinity(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_placement_policy_soft_anti_affinity.yaml',
            'hot_placement_policy_soft_anti_affinity.yaml', is_monitor=False)

    def test_create_tosca_with_placement_policy_default_affinity(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_placement_policy_default_affinity.yaml',
            'hot_placement_policy_default_affinity.yaml', is_monitor=False)

    def test_create_tosca_with_flavor_and_capabilities(self):
        self._test_assert_equal_for_tosca_templates(
            'test_tosca_flavor_and_capabilities.yaml',
            'hot_flavor_and_capabilities.yaml')

    def test_create_tosca_with_flavor_no_units(self):
        self._test_assert_equal_for_tosca_templates(
            'test_tosca_flavor_no_units.yaml',
            'hot_flavor_no_units.yaml')

    def test_create_tosca_with_flavor_extra_specs_all_numa_count(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_flavor_all_numa_count.yaml',
            'hot_tosca_flavor_all_numa_count.yaml')

    def test_create_tosca_with_flavor_extra_specs_all_numa_nodes(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_flavor_all_numa_nodes.yaml',
            'hot_tosca_flavor_all_numa_nodes.yaml')

    def test_create_tosca_with_flavor_extra_specs_numa_node_count_trumps(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_flavor_numa_nodes_count.yaml',
            'hot_tosca_flavor_numa_nodes_count.yaml')

    def test_create_tosca_with_flavor_extra_specs_huge_pages(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_flavor_huge_pages.yaml',
            'hot_tosca_flavor_huge_pages.yaml')

    def test_create_tosca_with_flavor_extra_specs_cpu_allocations(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_flavor_cpu_allocations.yaml',
            'hot_tosca_flavor_cpu_allocations.yaml')

    def test_create_tosca_with_flavor_extra_specs_numa_nodes(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_flavor_numa_nodes.yaml',
            'hot_tosca_flavor_numa_nodes.yaml')

    def test_create_tosca_with_new_image(self):
        self._test_assert_equal_for_tosca_templates('test_tosca_image.yaml',
            'hot_tosca_image.yaml')

    def test_create_tosca_sriov(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_sriov.yaml',
            'hot_tosca_sriov.yaml'
        )

    def test_create_tosca_vnic_normal(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_vnic_port.yaml',
            'hot_tosca_vnic_normal.yaml'
        )

    def test_create_tosca_mgmt_sriov_port(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_mgmt_sriov.yaml',
            'hot_tosca_mgmt_sriov.yaml'
        )

    def test_tosca_params(self):
        input_params = 'image: cirros\nflavor: m1.large'
        self._test_assert_equal_for_tosca_templates(
            'tosca_generic_vnfd_params.yaml',
            'hot_tosca_generic_vnfd_params.yaml',
            input_params
        )

    def test_tosca_artifacts_image_params(self):
        input_params = 'image_source: ' \
                       'http://download.cirros-cloud.net/' \
                       '0.5.2/cirros-0.5.2-x86_64-disk.img'
        self._test_assert_equal_for_tosca_templates(
            'tosca_artifacts_image_vnfd_params.yaml',
            'hot_tosca_artifacts_image_vnfd_params.yaml',
            input_params
        )

    def test_create_tosca_scale(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_scale.yaml',
            'hot_scale_main.yaml',
            files={'SP1_res.yaml': 'hot_scale_custom.yaml'},
            is_monitor=False
        )

    def test_get_resource_info(self):
        vnf_obj = self._get_expected_active_vnf()
        self.assertRaises(vnfm.InfraDriverUnreachable,
                          self.infra_driver.get_resource_info,
                          plugin=None, context=self.context, vnf_info=vnf_obj,
                          auth_attr=utils.get_vim_auth_obj(),
                          region_name=None)

    def test_create_port_with_security_groups(self):
        self._test_assert_equal_for_tosca_templates(
            'test_tosca_security_groups.yaml',
            'hot_tosca_security_groups.yaml'
        )

    def test_create_port_with_allowed_address_pairs(self):
        self._test_assert_equal_for_tosca_templates(
            'test_tosca_allowed_address_pairs.yaml',
            'hot_tosca_allowed_address_pairs.yaml'
        )

    def test_create_port_with_mac_and_ip(self):
        self._test_assert_equal_for_tosca_templates(
            'test_tosca_mac_ip.yaml',
            'hot_tosca_mac_ip.yaml'
        )

    def test_create_tosca_alarm_respawn(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_alarm_respawn.yaml',
            'hot_tosca_alarm_respawn.yaml',
            is_monitor=False
        )

    def test_create_tosca_alarm_scale(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_alarm_scale.yaml',
            'hot_tosca_alarm_scale.yaml',
            files={'SP1_res.yaml': 'hot_alarm_scale_custom.yaml'},
            is_monitor=False
        )

    def test_create_tosca_with_alarm_monitoring_not_matched(self):
        self.assertRaises(vnfm.MetadataNotMatched,
                          self._test_assert_equal_for_tosca_templates,
                          'tosca_alarm_metadata.yaml',
                          'hot_tosca_alarm_metadata.yaml',
                          is_monitor=False
                          )

    def test_create_tosca_monitoring_multi_vdus(self):
        self._test_assert_equal_for_tosca_templates(
            'tosca_monitoring_multi_vdu.yaml',
            'hot_tosca_monitoring_multi_vdu.yaml',
            multi_vdus=True
        )



# Copyright 2016 Brocade Communications System, Inc.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy import schema
from sqlalchemy import sql

from tacker.db import model_base
from tacker.db import models_v1
from tacker.db import types


class Vim(model_base.BASE,
          models_v1.HasId,
          models_v1.HasTenant,
          models_v1.Audit):
    type = sa.Column(sa.String(64), nullable=False)
    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.Text, nullable=True)
    placement_attr = sa.Column(types.Json, nullable=True)
    shared = sa.Column(sa.Boolean, default=False, server_default=sql.false(
    ), nullable=False)
    is_default = sa.Column(sa.Boolean, default=False, server_default=sql.false(
    ), nullable=False)
    vim_auth = orm.relationship('VimAuth')
    status = sa.Column(sa.String(255), nullable=False)
    extra = sa.Column(types.Json, nullable=True)

    __table_args__ = (
        schema.UniqueConstraint(
            "tenant_id",
            "name",
            "deleted_at",
            name="uniq_vim0tenant_id0name0deleted_at"),
    )


class VimAuth(model_base.BASE, models_v1.HasId):
    vim_id = sa.Column(types.Uuid, sa.ForeignKey('vims.id'),
                       nullable=False)
    password = sa.Column(sa.String(255), nullable=False)
    auth_url = sa.Column(sa.String(255), nullable=False)
    vim_project = sa.Column(types.Json, nullable=False)
    auth_cred = sa.Column(types.Json, nullable=False)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class SolException(Exception):
    """Exception for SOL ProblemDetails

    Generally status, title and message should be defined in derived class.
    detail is constructed from message and kwargs.

    Attributes in ProblemDetails can be specified in kwargs of object
    initialization. Use `sol_*` (ex. `sol_instance`) to avoid confliction
    with kwargs.
    """

    status = 500
    title = None
    message = 'Internal Server Error'

    def __init__(self, **kwargs):
        self.status = kwargs.pop('sol_status', self.status)
        self.title = kwargs.pop('sol_title', self.title)
        self.type = kwargs.pop('sol_type', None)
        self.instance = kwargs.pop('sol_instance', None)
        self.detail = kwargs.pop('sol_detail', self.message % kwargs)
        super().__init__(self.detail)

    def make_problem_details(self):
        res = {'status': self.status,
               'detail': self.detail}
        if self.title is not None:
            res['title'] = self.title
        if self.type is not None:
            res['type'] = self.type
        if self.instance is not None:
            res['instance'] = self.instance

        return res


class SolHttpError400(SolException):
    status = 400
    title = 'Bad Request'


class SolHttpError403(SolException):
    status = 403
    title = 'Forbidden'


class SolHttpError404(SolException):
    status = 404
    title = 'Not Found'


class SolHttpError405(SolException):
    status = 405
    title = 'Method Not Allowed'


class SolHttpError406(SolException):
    status = 406
    title = 'Not Acceptable'


class SolHttpError409(SolException):
    status = 409
    title = 'Conflict'


class SolHttpError422(SolException):
    status = 422
    title = 'Unprocessable Entity'


class MethodNotAllowed(SolHttpError405):
    message = _("Method %(method)s is not supported.")


class SolValidationError(SolHttpError400):
    message = _("%(detail)s")


class InvalidAPIVersionString(SolHttpError400):
    message = _("Version String %(version)s is of invalid format. Must "
                "be of format Major.Minor.Patch.")


class APIVersionMissing(SolHttpError400):
    message = _("'Version' HTTP header missing.")


class APIVersionNotSupported(SolHttpError406):
    message = _("Version %(version)s not supported.")


class VnfdIdNotEnabled(SolHttpError422):
    message = _("VnfId %(vnfd_id)s not ENABLED.")


class VnfInstanceNotFound(SolHttpError404):
    message = _("VnfInstance %(inst_id)s not found.")


class VnfInstanceIsInstantiated(SolHttpError409):
    message = _("VnfInstance %(inst_id)s is instantiated.")


class VnfInstanceIsNotInstantiated(SolHttpError409):
    message = _("VnfInstance %(inst_id)s isn't instantiated.")


class LccnSubscriptionNotFound(SolHttpError404):
    message = _("LccnSubscription %(subsc_id)s not found.")


class VnfLcmOpOccNotFound(SolHttpError404):
    message = _("VnfLcmOpOcc %(lcmocc_id)s not found.")


class VnfdIdNotFound(SolHttpError422):
    message = _("VnfPackage of vnfdId %(vnfd_id)s is not found or "
                "not operational.")


class FlavourIdNotFound(SolHttpError400):
    message = _("FlavourId %(flavour_id)s not found in the vnfd.")


class NoVimConnectionInfo(SolHttpError422):
    message = _("No VimConnectionInfo set to the VnfInstance.")


class InvalidVnfdFormat(SolHttpError400):
    message = _("Vnfd is unexpected format.")


class StackOperationFailed(SolHttpError422):
    # title and detail are set in the code from stack_status_reason
    pass


class MgmtDriverExecutionFailed(SolHttpError422):
    title = 'Mgmt driver execution failed'
    # detail set in the code


class BaseHOTNotDefined(SolHttpError400):
    message = _("BaseHOT is not defined.")


class UserdataMissing(SolHttpError400):
    message = _("'lcm-operation-user-data' or "
                "'lcm-operation-user-data-class' missing.")


class UserdataExecutionFailed(SolHttpError422):
    title = 'Userdata execution failed'
    # detail set in the code


class TestNotificationFailed(SolHttpError422):
    message = _("Can't get from notification callback Uri.")


class VimNotFound(SolHttpError404):
    message = _("VIM %(vim_id)s not found.")


class OtherOperationInProgress(SolHttpError409):
    message = _("Other LCM operation of vnfInstance %(inst_id)s "
                "is in progress.")


class UserDataClassNotImplemented(SolHttpError400):
    message = _("Userdata class not implemented.")


class InvalidAttributeFilter(SolHttpError400):
    message = _("Attribute filter expression is invalid.")


class InvalidAttributeSelector(SolHttpError400):
    message = _("Attribute selector expression is invalid.")


class InvalidSubscription(SolHttpError400):
    # detail set in the code
    pass


class ResponseTooBig(SolHttpError400):
    title = 'Response too big'
    message = _("Content length of the response is larger "
                "than %(size)d bytes.")


class LocalNfvoGrantFailed(SolHttpError403):
    title = 'Grant failed'
    # detail set in the code


class LcmOpOccNotFailedTemp(SolHttpError409):
    message = _("LCM operation %(lcmocc_id)s not FAILED_TEMP.")


class GrantRequestOrGrantNotFound(SolHttpError404):
    message = _("GrantRequest or Grant for LCM operation "
                "%(lcmocc_id)s not found.")


class RollbackNotSupported(SolHttpError422):
    message = _("Rollback of %(op)s is not supported.")


class UnexpectedParentResourceDefinition(SolHttpError422):
    message = _("Parent resource is necessary for VDU definition.")


class InvalidScaleAspectId(SolHttpError400):
    message = _("Invalid aspectId '%(aspect_id)s'.")


class InvalidScaleNumberOfSteps(SolHttpError400):
    message = _("Invalid numberOfSteps '%(num_steps)d'.")


class DeltaMissingInVnfd(SolHttpError400):
    message = _("Delta '%(delta)s' is not defined in "
                "VduScalingAspectDeltas.")


class ConductorProcessingError(SolException):
    title = 'Internal Server Error'
    message = _("Failure due to conductor processing error.")



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from datetime import datetime

from oslo_log import log as logging
from oslo_utils import uuidutils

from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import subscription_utils as subsc_utils
from tacker.sol_refactored.common import vnf_instance_utils as inst_utils
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects.v2 import fields


LOG = logging.getLogger(__name__)  # not used at the moment


def get_lcmocc(context, lcmocc_id):
    lcmocc = objects.VnfLcmOpOccV2.get_by_id(context, lcmocc_id)
    if lcmocc is None:
        raise sol_ex.VnfLcmOpOccNotFound(lcmocc_id=id)
    return lcmocc


def get_lcmocc_all(context):
    return objects.VnfLcmOpOccV2.get_all(context)


def lcmocc_href(lcmocc_id, endpoint):
    return "{}/vnflcm/v2/vnf_lcm_op_occs/{}".format(endpoint, lcmocc_id)


def lcmocc_task_href(lcmocc_id, task, endpoint):
    return "{}/vnflcm/v2/vnf_lcm_op_occs/{}/{}".format(endpoint, lcmocc_id,
                                                       task)


def make_lcmocc_links(lcmocc, endpoint):
    links = objects.VnfLcmOpOccV2_Links()
    links.self = objects.Link(href=lcmocc_href(lcmocc.id, endpoint))
    links.vnfInstance = objects.Link(
        href=inst_utils.inst_href(lcmocc.vnfInstanceId, endpoint))
    links.retry = objects.Link(
        href=lcmocc_task_href(lcmocc.id, 'retry', endpoint))
    links.rollback = objects.Link(
        href=lcmocc_task_href(lcmocc.id, 'rollback', endpoint))
    links.fail = objects.Link(
        href=lcmocc_task_href(lcmocc.id, 'fail', endpoint))
    # TODO(oda-g): add when implemented
    # links.grant
    # links.cancel
    # links.vnfSnapshot

    return links


def make_lcmocc_notif_data(subsc, lcmocc, endpoint):
    notif_data = objects.VnfLcmOperationOccurrenceNotificationV2(
        id=uuidutils.generate_uuid(),
        notificationType="VnfLcmOperationOccurrenceNotification",
        subscriptionId=subsc.id,
        timeStamp=datetime.utcnow(),
        operationState=lcmocc.operationState,
        vnfInstanceId=lcmocc.vnfInstanceId,
        operation=lcmocc.operation,
        isAutomaticInvocation=lcmocc.isAutomaticInvocation,
        verbosity=subsc.verbosity,
        vnfLcmOpOccId=lcmocc.id,
        _links=objects.LccnLinksV2(
            vnfInstance=objects.NotificationLink(
                href=inst_utils.inst_href(lcmocc.vnfInstanceId, endpoint)),
            subscription=objects.NotificationLink(
                href=subsc_utils.subsc_href(subsc.id, endpoint)),
            vnfLcmOpOcc=objects.NotificationLink(
                href=lcmocc_href(lcmocc.id, endpoint))
        )
    )

    if ((lcmocc.operation == fields.LcmOperationType.MODIFY_INFO and
         lcmocc.operationState == fields.LcmOperationStateType.PROCESSING) or
            lcmocc.operationState == fields.LcmOperationStateType.STARTING):
        notif_data.notificationStatus = 'START'
    else:
        notif_data.notificationStatus = 'RESULT'

    if lcmocc.obj_attr_is_set('error'):
        notif_data.error = lcmocc.error

    if notif_data.verbosity == fields.LcmOpOccNotificationVerbosityType.FULL:
        if lcmocc.obj_attr_is_set('resourceChanges'):
            attrs = ['affectedVnfcs',
                     'affectedVirtualLinks',
                     'affectedExtLinkPorts',
                     'affectedVirtualStorages']
            for attr in attrs:
                if lcmocc.resourceChanges.obj_attr_is_set(attr):
                    notif_data[attr] = lcmocc.resourceChanges[attr]
        attrs = ['changedInfo',
                 'changedExtConnectivity',
                 'modificationsTriggeredByVnfPkgChange']
        for attr in attrs:
            if lcmocc.obj_attr_is_set(attr):
                notif_data[attr] = lcmocc[attr]

    return notif_data


def _make_affected_vnfc(vnfc, change_type, strgs):
    affected_vnfc = objects.AffectedVnfcV2(
        id=vnfc.id,
        vduId=vnfc.vduId,
        changeType=change_type,
        computeResource=vnfc.computeResource
    )
    if vnfc.obj_attr_is_set('vnfcCpInfo'):
        cp_ids = [cp.id for cp in vnfc.vnfcCpInfo]
        affected_vnfc.affectedVnfcCpIds = cp_ids
    if vnfc.obj_attr_is_set('storageResourceIds'):
        # NOTE: in case of heal, volume may not be deleted/re-created.
        if change_type == 'ADDED':
            str_ids = [str_id for str_id in vnfc.storageResourceIds
                       if str_id in strgs]
            if str_ids:
                affected_vnfc.addedStorageResourceIds = str_ids
        else:  # 'REMOVED'
            str_ids = [str_id for str_id in vnfc.storageResourceIds
                       if str_id in strgs]
            if str_ids:
                affected_vnfc.removedStorageResourceIds = str_ids

    return affected_vnfc


def _make_affected_vl(vl, change_type):
    affected_vl = objects.AffectedVirtualLinkV2(
        id=vl.id,
        vnfVirtualLinkDescId=vl.vnfVirtualLinkDescId,
        changeType=change_type,
        networkResource=vl.networkResource
    )
    if vl.obj_attr_is_set('vnfLinkPorts'):
        affected_vl.vnfLinkPortIds = [port.id for port in vl.vnfLinkPorts]

    return affected_vl


def _make_affected_vls_link_port_change(vls_saved, vls, common_vls):
    affected_vls = []

    for vl_id in common_vls:
        old_ports = set()
        new_ports = set()
        for vl in vls_saved:
            if vl.id == vl_id:
                old_vl = vl
                if vl.obj_attr_is_set('vnfLinkPorts'):
                    old_ports = {port.id for port in vl.vnfLinkPorts}
        for vl in vls:
            if vl.id == vl_id:
                new_vl = vl
                if vl.obj_attr_is_set('vnfLinkPorts'):
                    new_ports = {port.id for port in vl.vnfLinkPorts}
        add_ports = new_ports - old_ports
        rm_ports = old_ports - new_ports
        # NOTE: Only for extManagedVirtualLink in case of heal
        # (SOL003 all=true) there may be both of add_ports and rm_ports.
        if add_ports:
            affected_vl = objects.AffectedVirtualLinkV2(
                id=new_vl.id,
                vnfVirtualLinkDescId=new_vl.vnfVirtualLinkDescId,
                changeType='LINK_PORT_ADDED',
                networkResource=new_vl.networkResource,
                vnfLinkPortIds=list(add_ports)
            )
            affected_vls.append(affected_vl)
        if rm_ports:
            affected_vl = objects.AffectedVirtualLinkV2(
                id=old_vl.id,
                vnfVirtualLinkDescId=old_vl.vnfVirtualLinkDescId,
                changeType='LINK_PORT_REMOVED',
                networkResource=old_vl.networkResource,
                vnfLinkPortIds=list(rm_ports)
            )
            affected_vls.append(affected_vl)

    return affected_vls


def _make_affected_strg(strg, change_type):
    return objects.AffectedVirtualStorageV2(
        id=strg.id,
        virtualStorageDescId=strg.virtualStorageDescId,
        changeType=change_type,
        storageResource=strg.storageResource
    )


def _make_affected_ext_link_ports(inst_info_saved, inst_info):
    affected_ext_link_ports = []

    ext_vl_ports_saved = set()
    ext_vl_ports = set()
    if inst_info_saved.obj_attr_is_set('extVirtualLinkInfo'):
        for ext_vl in inst_info_saved.extVirtualLinkInfo:
            if ext_vl.obj_attr_is_set('extLinkPorts'):
                ext_vl_ports_saved |= {port.id
                    for port in ext_vl.extLinkPorts}
    if inst_info.obj_attr_is_set('extVirtualLinkInfo'):
        for ext_vl in inst_info.extVirtualLinkInfo:
            if ext_vl.obj_attr_is_set('extLinkPorts'):
                ext_vl_ports |= {port.id
                    for port in ext_vl.extLinkPorts}
    add_ext_vl_ports = ext_vl_ports - ext_vl_ports_saved
    rm_ext_vl_ports = ext_vl_ports_saved - ext_vl_ports

    if add_ext_vl_ports:
        for ext_vl in inst_info.extVirtualLinkInfo:
            if not ext_vl.obj_attr_is_set('extLinkPorts'):
                continue
            affected_ext_link_ports += [
                objects.AffectedExtLinkPortV2(
                    id=port.id,
                    changeType='ADDED',
                    extCpInstanceId=port.cpInstanceId,
                    resourceHandle=port.resourceHandle
                )
                for port in ext_vl.extLinkPorts
                if port.id in add_ext_vl_ports
            ]
    if rm_ext_vl_ports:
        for ext_vl in inst_info_saved.extVirtualLinkInfo:
            if not ext_vl.obj_attr_is_set('extLinkPorts'):
                continue
            affected_ext_link_ports += [
                objects.AffectedExtLinkPortV2(
                    id=port.id,
                    changeType='REMOVED',
                    extCpInstanceId=port.cpInstanceId,
                    resourceHandle=port.resourceHandle
                )
                for port in ext_vl.extLinkPorts
                if port.id in rm_ext_vl_ports
            ]

    return affected_ext_link_ports


def _check_modification(inst_saved, inst, attr):
    if not inst.obj_attr_is_set(attr):
        return False
    if (not inst_saved.obj_attr_is_set(attr) or
            getattr(inst_saved, attr) != getattr(inst, attr)):
        return True
    return False


def _change_vnf_info(lcmocc, inst_saved, inst):
    vnf_info_modify = objects.VnfInfoModificationsV2()

    # vnfdid is required, don't check the existence of the value.
    if inst_saved.vnfdId != inst.vnfdId:
        vnf_info_modify.vnfdId = inst.vnfdId

        if inst_saved.vnfProvider != inst.vnfProvider:
            vnf_info_modify.vnfProvider = inst.vnfProvider
        if inst_saved.vnfProductName != inst.vnfProductName:
            vnf_info_modify.vnfProductName = inst.vnfProductName
        if inst_saved.vnfSoftwareVersion != inst.vnfSoftwareVersion:
            vnf_info_modify.vnfSoftwareVersion = inst.vnfSoftwareVersion
        if inst_saved.vnfdVersion != inst.vnfdVersion:
            vnf_info_modify.vnfdVersion = inst.vnfdVersion

    attrs = ['vnfInstanceName', 'vnfInstanceDescription',
             'vnfConfigurableProperties', 'metadata', 'extensions']
    for attr in attrs:
        if _check_modification(inst_saved, inst, attr):
            setattr(vnf_info_modify, attr, getattr(inst, attr))

    if (inst.obj_attr_is_set('vimConnectionInfo') and
            inst_saved.obj_attr_is_set('vimConnectionInfo')):
        inst_viminfo = inst.to_dict()["vimConnectionInfo"]
        inst_saved_viminfo = inst_saved.to_dict()["vimConnectionInfo"]
        if inst_viminfo != inst_saved_viminfo:
            vnf_info_modify.vimConnectionInfo = inst.vimConnectionInfo
    elif (inst.obj_attr_is_set('vimConnectionInfo') and
            not inst_saved.obj_attr_is_set('vimConnectionInfo')):
        vnf_info_modify.vimConnectionInfo = inst.vimConnectionInfo

    vnfc_info_mod = []
    vnfc_info = []
    if (inst.obj_attr_is_set('instantiatedVnfInfo') and
            inst.instantiatedVnfInfo.obj_attr_is_set('vnfcInfo')):
        vnfc_info = inst.instantiatedVnfInfo.vnfcInfo
    vnfc_info_saved = []
    if (inst_saved.obj_attr_is_set('instantiatedVnfInfo') and
            inst_saved.instantiatedVnfInfo.obj_attr_is_set('vnfcInfo')):
        vnfc_info_saved = inst_saved.instantiatedVnfInfo.vnfcInfo

    for vnfc in vnfc_info:
        prop_saved = {}
        for vnfc_saved in vnfc_info_saved:
            if vnfc.id == vnfc_saved.id:
                if vnfc_saved.obj_attr_is_set('vnfcConfigurableProperties'):
                    prop_saved = vnfc_saved.vnfcConfigurableProperties
                break
        prop = {}
        if vnfc.obj_attr_is_set('vnfcConfigurableProperties'):
            prop = vnfc.vnfcConfigurableProperties
        if prop != prop_saved:
            vnfc_info_mod.append(
                objects.VnfcInfoModificationsV2(
                    id=vnfc.id,
                    vnfcConfigurableProperties=prop
                )
            )
    if vnfc_info_mod:
        vnf_info_modify.vnfcInfoModifications = vnfc_info_mod

    lcmocc.changedInfo = vnf_info_modify


def update_lcmocc(lcmocc, inst_saved, inst):
    # if operation is MODIFY_INFO, make changedInfo of lcmocc.
    # for other operations, make ResourceChanges of lcmocc
    # from instantiatedVnfInfo. In addition if operation is
    # CHANGE_EXT_CONN, make changedExtConnectivity of lcmocc.
    # NOTE: grant related info such as resourceDefinitionId, zoneId
    # and so on are not included in lcmocc since such info are not
    # included in instantiatedVnfInfo.

    if lcmocc.operation == fields.LcmOperationType.MODIFY_INFO:
        _change_vnf_info(lcmocc, inst_saved, inst)
        return

    if inst_saved.obj_attr_is_set('instantiatedVnfInfo'):
        inst_info_saved = inst_saved.instantiatedVnfInfo
    else:
        # dummy
        inst_info_saved = objects.VnfInstanceV2_InstantiatedVnfInfo()

    inst_info = inst.instantiatedVnfInfo

    # NOTE: objects may be re-created. so compare 'id' instead of object
    # itself.
    def _calc_diff(attr):
        # NOTE: instantiatedVnfInfo object is dict compat
        objs_saved = set()
        if inst_info_saved.obj_attr_is_set(attr):
            objs_saved = {obj.id for obj in inst_info_saved[attr]}
        objs = set()
        if inst_info.obj_attr_is_set(attr):
            objs = {obj.id for obj in inst_info[attr]}

        # return removed_objs, added_objs, common_objs
        return objs_saved - objs, objs - objs_saved, objs_saved & objs

    removed_strgs, added_strgs, _ = _calc_diff('virtualStorageResourceInfo')
    affected_strgs = []
    if removed_strgs:
        affected_strgs += [
            _make_affected_strg(strg, 'REMOVED')
            for strg in inst_info_saved.virtualStorageResourceInfo
            if strg.id in removed_strgs
        ]
    if added_strgs:
        affected_strgs += [_make_affected_strg(strg, 'ADDED')
                           for strg in inst_info.virtualStorageResourceInfo
                           if strg.id in added_strgs]

    removed_vnfcs, added_vnfcs, _ = _calc_diff('vnfcResourceInfo')
    affected_vnfcs = []
    if removed_vnfcs:
        affected_vnfcs += [
            _make_affected_vnfc(vnfc, 'REMOVED', removed_strgs)
            for vnfc in inst_info_saved.vnfcResourceInfo
            if vnfc.id in removed_vnfcs
        ]
    if added_vnfcs:
        affected_vnfcs += [
            _make_affected_vnfc(vnfc, 'ADDED', added_strgs)
            for vnfc in inst_info.vnfcResourceInfo
            if vnfc.id in added_vnfcs
        ]

    removed_vls, added_vls, common_vls = _calc_diff(
        'vnfVirtualLinkResourceInfo')
    affected_vls = []
    if removed_vls:
        affected_vls += [_make_affected_vl(vl, 'REMOVED')
                         for vl in inst_info_saved.vnfVirtualLinkResourceInfo
                         if vl.id in removed_vls]
    if added_vls:
        affected_vls += [_make_affected_vl(vl, 'ADDED')
                         for vl in inst_info.vnfVirtualLinkResourceInfo
                         if vl.id in added_vls]
    if common_vls:
        affected_vls += _make_affected_vls_link_port_change(
            inst_info_saved.vnfVirtualLinkResourceInfo,
            inst_info.vnfVirtualLinkResourceInfo, common_vls)

    removed_mgd_vls, added_mgd_vls, common_mgd_vls = _calc_diff(
        'extManagedVirtualLinkInfo')
    if removed_mgd_vls:
        affected_vls += [_make_affected_vl(vl, 'LINK_PORT_REMOVED')
                         for vl in inst_info_saved.extManagedVirtualLinkInfo
                         if vl.id in removed_mgd_vls]
    if added_mgd_vls:
        affected_vls += [_make_affected_vl(vl, 'LINK_PORT_ADDED')
                         for vl in inst_info.extManagedVirtualLinkInfo
                         if vl.id in added_mgd_vls]
    if common_mgd_vls:
        affected_vls += _make_affected_vls_link_port_change(
            inst_info_saved.extManagedVirtualLinkInfo,
            inst_info.extManagedVirtualLinkInfo, common_mgd_vls)

    affected_ext_link_ports = _make_affected_ext_link_ports(
        inst_info_saved, inst_info)

    if (affected_vnfcs or affected_vls or affected_strgs or
            affected_ext_link_ports):
        change_info = objects.VnfLcmOpOccV2_ResourceChanges()
        if affected_vnfcs:
            change_info.affectedVnfcs = affected_vnfcs
        if affected_vls:
            change_info.affectedVirtualLinks = affected_vls
        if affected_strgs:
            change_info.affectedVirtualStorages = affected_strgs
        if affected_ext_link_ports:
            change_info.affectedExtLinkPorts = affected_ext_link_ports
        lcmocc.resourceChanges = change_info

    if lcmocc.operation == fields.LcmOperationType.CHANGE_EXT_CONN:
        _, added_ext_vls, common_ext_vls = _calc_diff('extVirtualLinkInfo')

        def _get_ext_vl(vl_id, vl_array):
            for vl in vl_array:
                if vl.id == vl_id:
                    return vl

        chg_ext_conn = [_get_ext_vl(ext_vl_id, inst_info.extVirtualLinkInfo)
                        for ext_vl_id in added_ext_vls]

        for ext_vl_id in common_ext_vls:
            ext_vl = _get_ext_vl(ext_vl_id, inst_info.extVirtualLinkInfo)
            ext_vl_saved = _get_ext_vl(ext_vl_id,
                                       inst_info_saved.extVirtualLinkInfo)
            cp_data = []
            if ext_vl.obj_attr_is_set('currentVnfExtCpData'):
                cp_data = sorted(ext_vl.to_dict()['currentVnfExtCpData'],
                                 key=lambda x: x['cpdId'])
            cp_data_saved = []
            if ext_vl_saved.obj_attr_is_set('currentVnfExtCpData'):
                cp_data_saved = sorted(
                    ext_vl_saved.to_dict()['currentVnfExtCpData'],
                    key=lambda x: x['cpdId'])
            if cp_data != cp_data_saved:
                chg_ext_conn.append(ext_vl)
                continue

            # NOTE: For ports created by the heat, the id is not changed if
            # its resourceId is not changed. But for ports given outside the
            # heat, the resourceId may be changed without changing the id,
            # so it is a policy to set changedExtConnectivity when there is
            # a change in "id" or "resourceHandle.resourceId".
            port_ids = set()
            if ext_vl.obj_attr_is_set('extLinkPorts'):
                port_ids = {(port.id, port.resourceHandle.resourceId)
                            for port in ext_vl.extLinkPorts}
            port_ids_saved = set()
            if ext_vl_saved.obj_attr_is_set('extLinkPorts'):
                port_ids_saved = {(port.id, port.resourceHandle.resourceId)
                                  for port in ext_vl_saved.extLinkPorts}
            if port_ids != port_ids_saved:
                chg_ext_conn.append(ext_vl)

        if chg_ext_conn:
            lcmocc.changedExtConnectivity = chg_ext_conn


def get_grant_req_and_grant(context, lcmocc):
    if lcmocc.operation == fields.LcmOperationType.MODIFY_INFO:
        return None, None

    grant_reqs = objects.GrantRequestV1.get_by_filter(context,
                                                      vnfLcmOpOccId=lcmocc.id)
    grant = objects.GrantV1.get_by_id(context, lcmocc.grantId)
    if not grant_reqs or grant is None:
        raise sol_ex.GrantRequestOrGrantNotFound(lcmocc_id=lcmocc.id)

    # len(grant_reqs) == 1 because vnfLcmOpOccId is primary key.
    return grant_reqs[0], grant


def check_lcmocc_in_progress(context, inst_id):
    # if the controller or conductor executes an operation for the vnf
    # instance (i.e. operationState is ...ING), other operation for
    # the same vnf instance is excluded by the coordinator.
    # check here is existence of lcmocc for the vnf instance with
    # FAILED_TEMP operationState.
    lcmoccs = objects.VnfLcmOpOccV2.get_by_filter(
        context, vnfInstanceId=inst_id,
        operationState=fields.LcmOperationStateType.FAILED_TEMP)
    if lcmoccs:
        raise sol_ex.OtherOperationInProgress(inst_id=inst_id)



# Copyright (c) 2014-2018 China Mobile (SuZhou) Software Technology Co.,Ltd.
# All Rights Reserved
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
from unittest import mock

from tacker.common import clients
from tacker import context
from tacker.db.common_services import common_services_db_plugin
from tacker.plugins.common import constants
from tacker.vnfm.infra_drivers.openstack import heat_client as hc
from tacker.vnfm.policy_actions.respawn import respawn as \
    policy_actions_respawn
from tacker.vnfm import vim_client


class VNFActionRespawn(testtools.TestCase):

    def setUp(self):
        super(VNFActionRespawn, self).setUp()
        self.context = context.get_admin_context()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()

    @mock.patch.object(clients.OpenstackClients, 'heat')
    @mock.patch.object(hc.HeatClient, 'delete')
    @mock.patch.object(vim_client.VimClient, 'get_vim')
    def test_execute_action(self, mock_get_vim, mock_hc_delete, mock_heat):
        action_respawn = policy_actions_respawn.VNFActionRespawn()
        vnf_dict = {
            'id': 'fake-id',
            'status': 'fake-status',
            'attributes': {
                'monitoring_policy': 'fake-monitoring-policy',
                'failure_count': '1',
                'dead_instance_id_1': '00000000-0000-0000-0000-00000000001'},
            'vim_id': 'fake-vim-id',
            'vim_auth': 'fake-vim-auth',
            'instance_id': '00000000-0000-0000-0000-000000000002',
            'placement_attr': {
                'region_name': 'fake-region-name'}}
        mock_get_vim.return_value = {'vim_auth': {
            'auth_url': 'http://fake-url/identity/v3'
        }}
        mock_hc_delete.return_value = True
        plugin = mock.Mock()
        plugin._mark_vnf_dead.return_value = True
        plugin.create_vnf_sync.return_value = {'id': 'fake-id'}
        plugin._vnf_monitor = mock.Mock()
        action_respawn.execute_action(plugin, self.context, vnf_dict, None)
        self._cos_db_plugin.create_event.assert_called_once_with(
            self.context, res_id=vnf_dict['id'],
            res_state=vnf_dict['status'],
            res_type=constants.RES_TYPE_VNF,
            evt_type=constants.RES_EVT_MONITOR,
            tstamp=mock.ANY, details="ActionRespawnHeat invoked")
        mock_get_vim.assert_called_once_with(self.context, vnf_dict['vim_id'])
        plugin.create_vnf_sync.assert_called_with(self.context, vnf_dict)
        plugin._vnf_monitor.mark_dead.assert_called_once_with(vnf_dict['id'])



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import collections
import contextlib
import datetime

from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_serialization import jsonutils
from oslo_utils import versionutils
from oslo_versionedobjects import base as ovoo_base
from oslo_versionedobjects import exception as ovoo_exc

from tacker.db import api as db_api
from tacker.sol_refactored.db.sqlalchemy import models
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects import fields as obj_fields


LOG = logging.getLogger(__name__)


def get_attrname(name):
    """Return the mangled name of the attribute's underlying storage."""
    return '_obj_' + name


# Every TackerObject has its attributes named exactly same as the corresponding
# ETSI NFV-SOL data type attributes. In almost all cases, so as its one-to-one
# mapped model attributes and DB column names. However, some field names
# cannot be used for model attributes (e.g., 'metadata'), in which case we
# append a suffix '__' (double underscores) as a workaround.
#
# Note that TackerObject users need not to concern about this.
#
# Example:
# - NFV-SOL data type "VnfXXX"
#     - id
#     - name
#     - metadata
# - class VnfXXX(base.TackerPersistentObject) attributes
#     - id
#     - name
#     - metadata
# - class VnfXXX(model_base.BASE) attributes
#     - id
#     - name
#     - metadata__  <-- 'metadata' is one of reserved names and cannot be used.
# - DB column names:
#     - id
#     - name
#     - metadata
RESERVED_FIELD_NAMES = [
    'metadata',
]


def get_model_field(name):
    if name in RESERVED_FIELD_NAMES:
        return name + '__'
    return name


class TackerObjectRegistry(ovoo_base.VersionedObjectRegistry):
    notification_classes = []
    _registry = None

    # NOTE: Separate registry from original tacker objects.
    # When original tacker objects are gone, _registry and
    # __new__ can be removed.
    def __new__(cls, *args, **kwargs):
        if not TackerObjectRegistry._registry:
            TackerObjectRegistry._registry = object.__new__(
                TackerObjectRegistry, *args, **kwargs)
            TackerObjectRegistry._registry._obj_classes = \
                collections.defaultdict(list)
        self = object.__new__(cls, *args, **kwargs)
        self._obj_classes = TackerObjectRegistry._registry._obj_classes
        return self

    def registration_hook(self, cls, index):
        # NOTE: This is called when an object is registered,
        # and is responsible for maintaining tacker.objects.$OBJECT
        # as the highest-versioned implementation of a given object.
        version = versionutils.convert_version_to_tuple(cls.VERSION)
        if not hasattr(objects, cls.obj_name()):
            setattr(objects, cls.obj_name(), cls)
        else:
            cur_version = versionutils.convert_version_to_tuple(
                getattr(objects, cls.obj_name()).VERSION)
            if version >= cur_version:
                setattr(objects, cls.obj_name(), cls)


class TackerObject(ovoo_base.VersionedObject):
    # These should be renamed (s/_sol_refactored//) when the older
    # implementation ceases to exist.
    OBJ_SERIAL_NAMESPACE = 'tacker_sol_refactored_object'
    OBJ_PROJECT_NAMESPACE = 'tacker_sol_refactored'

    def __init__(self, context=None, **kwargs):
        super(TackerObject, self).__init__(context, **kwargs)
        self.obj_set_defaults()

    def tacker_obj_get_changes(self):
        """Returns a dict of changed fields with tz unaware datetimes.

        Any timezone aware datetime field will be converted to UTC timezone
        and returned as timezone unaware datetime.

        This will allow us to pass these fields directly to a db update
        method as they can't have timezone information.
        """
        # Get dirtied/changed fields
        changes = self.obj_get_changes()

        # Look for datetime objects that contain timezone information
        for k, v in changes.items():
            if isinstance(v, datetime.datetime) and v.tzinfo:
                # Remove timezone information and adjust the time according to
                # the timezone information's offset.
                changes[k] = v.replace(tzinfo=None) - v.utcoffset()

        # Return modified dict
        return changes

    def obj_reset_changes(self, fields=None, recursive=False):
        """Reset the list of fields that have been changed.

        .. note::

          - This is NOT "revert to previous values"
          - Specifying fields on recursive resets will only be honored at the
            top level. Everything below the top will reset all.

        :param fields: List of fields to reset, or "all" if None.
        :param recursive: Call obj_reset_changes(recursive=True) on
                          any sub-objects within the list of fields
                          being reset.
        """
        if recursive:
            for field in self.obj_get_changes():

                # Ignore fields not in requested set (if applicable)
                if fields and field not in fields:
                    continue

                # Skip any fields that are unset
                if not self.obj_attr_is_set(field):
                    continue

                value = getattr(self, field)

                # Don't reset nulled fields
                if value is None:
                    continue

                # Reset straight Object and ListOfObjects fields
                if isinstance(self.fields[field], obj_fields.ObjectField):
                    value.obj_reset_changes(recursive=True)
                elif isinstance(self.fields[field],
                                obj_fields.ListOfObjectsField):
                    for thing in value:
                        thing.obj_reset_changes(recursive=True)
                elif isinstance(self.fields[field],
                                obj_fields.DictOfObjectsField):
                    for key, thing in value.items():
                        thing.obj_reset_changes(recursive=True)

        if fields:
            self._changed_fields -= set(fields)
        else:
            self._changed_fields.clear()

    @classmethod
    def from_dict(cls, dict_obj):
        inst = cls()
        for name, field in cls.fields.items():
            value = dict_obj.get(name, None)
            if value is None:
                continue
            if isinstance(field, obj_fields.ObjectField):
                child_cls = cls.obj_class_from_name(field.objname, cls.VERSION)
                setattr(inst, name, child_cls.from_dict(value))
            elif isinstance(field, obj_fields.ListOfObjectsField):
                child_cls = cls.obj_class_from_name(field.objname, cls.VERSION)
                list_of_objects = []
                for thing in value:
                    list_of_objects.append(child_cls.from_dict(thing))
                setattr(inst, name, list_of_objects)
            elif isinstance(field, obj_fields.DictOfObjectsField):
                child_cls = cls.obj_class_from_name(field.objname, cls.VERSION)
                dict_of_objects = {}
                for key, thing in value.items():
                    dict_of_objects[key] = child_cls.from_dict(thing)
                setattr(inst, name, dict_of_objects)
            else:
                setattr(inst, name, field.from_primitive(inst, name, value))
        return inst

    @classmethod
    def from_json(cls, json_obj):
        return cls.from_dict(jsonutils.loads(json_obj))

    def to_dict(self):
        obj = {}
        for name, field in self.fields.items():
            if not self.obj_attr_is_set(name):
                continue
            if getattr(self, name) is None:
                continue
            if isinstance(field, obj_fields.ObjectField):
                obj[name] = getattr(self, name).to_dict()
            elif isinstance(field, obj_fields.ListOfObjectsField):
                obj[name] = []
                for item in getattr(self, name):
                    obj[name].append(item.to_dict())
            elif isinstance(field, obj_fields.DictOfObjectsField):
                obj[name] = {}
                for key, item in getattr(self, name).items():
                    obj[name][key] = item.to_dict()
            else:
                obj[name] = field.to_primitive(self, name, getattr(self, name))
        return obj

    def to_json(self):
        return jsonutils.dumps(self.to_dict())

    @contextlib.contextmanager
    def obj_alternate_context(self, context):
        original_context = self._context
        self._context = context
        try:
            yield
        finally:
            self._context = original_context

    @classmethod
    def obj_class_from_name(cls, objname, objver):
        """Returns a class from the registry based on a name and version."""
        if objname not in TackerObjectRegistry.obj_classes():
            LOG.error('Unable to instantiate unregistered object type '
                      '%(objtype)s'), dict(objtype=objname)
            raise ovoo_exc.UnsupportedObjectError(objtype=objname)

        # NOTE: only the newest version is registered by registration_hook.
        # omit version check.
        return TackerObjectRegistry.obj_classes()[objname][0]


class TackerObjectSerializer(messaging.NoOpSerializer):
    """A TackerObject-aware Serializer.

    This implements the Oslo Serializer interface and provides the
    ability to serialize and deserialize TackerObject entities. Any service
    that needs to accept or return TackerObjects as arguments or result values
    should pass this to its RPCClient and RPCServer objects.
    """

    def _process_object(self, context, objprim):
        return TackerObject.obj_from_primitive(objprim, context=context)

    def _process_iterable(self, context, action_fn, values):
        """Process an iterable, taking an action on each value.

        :param:context: Request context
        :param:action_fn: Action to take on each item in values
        :param:values: Iterable container of things to take action on
        :returns: A new container of the same type (except set) with
                  items from values having had action applied.
        """

        iterable = values.__class__
        if issubclass(iterable, dict):
            return iterable(**{k: action_fn(context, v)
                               for k, v in values.items()})

        # NOTE: A set can't have an unhashable value inside,
        # such as a dict. Convert the set to list, which is fine, since we
        # can't send them over RPC anyway. We convert it to list as this
        # way there will be no semantic change between the fake rpc driver
        # used in functional test and a normal rpc driver.
        if iterable == set:
            iterable = list
        return iterable([action_fn(context, value) for value in values])

    def serialize_entity(self, context, entity):
        if isinstance(entity, (tuple, list, set, dict)):
            entity = self._process_iterable(context, self.serialize_entity,
                                            entity)
        elif (hasattr(entity, 'obj_to_primitive') and
              callable(entity.obj_to_primitive)):
            entity = entity.obj_to_primitive()
        return entity

    def deserialize_entity(self, context, entity):
        if (isinstance(entity, dict) and
                TackerObject._obj_primitive_key('name') in entity):
            entity = self._process_object(context, entity)
        elif isinstance(entity, (tuple, list, set, dict)):
            entity = self._process_iterable(context, self.deserialize_entity,
                                            entity)
        return entity


class TackerPersistentObject(TackerObject):
    """Class for objects supposed to be to DB."""

    def __init__(self, context=None, **kwargs):
        super(TackerPersistentObject, self).__init__(context, **kwargs)
        self._db_obj = None

    # By default, it's assumed that there is a model class corresponding to one
    # TackerPersistentObject, which has the same named fields.
    def _get_model_cls(self):
        clsname = self.__class__.__name__
        return getattr(models, clsname)

    @db_api.context_manager.writer
    def _save(self, context, merge=False):
        if not self.obj_get_changes():
            return
        model_cls = self._get_model_cls()
        inst = model_cls()
        inst.update(self.to_db_obj())
        # note: The same workaround is present in oslo.db ModelBase.save()
        #       implementation.
        with context.session.begin(subtransactions=True):
            if merge:
                context.session.merge(inst, load=True)
            else:
                context.session.add(inst)
            context.session.flush()
        # 'flush' must have succeeded because we are here.
        if self._db_obj is None:
            self._db_obj = inst
        self.obj_reset_changes()

    @db_api.context_manager.writer
    def delete(self, context):
        if self._db_obj is None:
            return
        context.session.delete(self._db_obj)

    # WARNING: Check if it is really necessary if you consider overriding this.
    def create(self, context):
        self._save(context)

    # WARNING: Check if it is really necessary if you consider overriding this.
    def update(self, context):
        self._save(context, merge=True)

    @classmethod
    @db_api.context_manager.reader
    def get_by_id(cls, context, id):
        model_cls = getattr(models, cls.__name__)
        query = context.session.query(model_cls).filter(model_cls.id == id)
        result = query.one_or_none()
        if result is None:
            return None
        return cls.from_db_obj(result)

    @classmethod
    @db_api.context_manager.reader
    def get_all(cls, context):
        model_cls = getattr(models, cls.__name__)
        query = context.session.query(model_cls)
        result = query.all()
        return [cls.from_db_obj(item) for item in result]

    @classmethod
    @db_api.context_manager.reader
    def get_by_filter(cls, context, *args, **kwargs):
        model_cls = getattr(models, cls.__name__)
        query = context.session.query(model_cls).filter_by(**kwargs)
        result = query.all()
        return [cls.from_db_obj(item) for item in result]

    @classmethod
    def from_db_obj(cls, db_obj):
        inst = cls()
        for name, field in cls.fields.items():
            name_ = get_model_field(name)
            value = db_obj.get(name_, None)
            if value is None:
                continue
            if isinstance(field, obj_fields.ObjectField):
                child_cls = cls.obj_class_from_name(field.objname, None)
                setattr(inst, name, child_cls.from_json(value))
            elif isinstance(field, obj_fields.ListOfObjectsField):
                child_cls = cls.obj_class_from_name(field.objname, None)
                list_of_objects = []
                value_loaded = jsonutils.loads(value)
                for thing in value_loaded:
                    list_of_objects.append(child_cls.from_dict(thing))
                setattr(inst, name, list_of_objects)
            elif isinstance(field, obj_fields.DictOfObjectsField):
                child_cls = cls.obj_class_from_name(field.objname, None)
                dict_of_objects = {}
                value_loaded = jsonutils.loads(value)
                for key, thing in value_loaded.items():
                    dict_of_objects[key] = child_cls.from_dict(thing)
                setattr(inst, name, dict_of_objects)
            elif isinstance(field, obj_fields.DateTimeField):
                setattr(inst, name, value)
            else:
                setattr(inst, name, field.from_primitive(inst, name, value))
        inst._db_obj = db_obj
        inst.obj_reset_changes()
        return inst

    def to_db_obj(self):
        obj = {}
        for name, field in self.fields.items():
            name_ = get_model_field(name)
            if not self.obj_attr_is_set(name):
                continue
            if getattr(self, name) is None:
                obj[name_] = None
                continue
            if isinstance(field, obj_fields.ObjectField):
                obj[name_] = getattr(self, name).to_json()
            elif isinstance(field, obj_fields.ListOfObjectsField):
                list_of_objects = []
                for item in getattr(self, name):
                    list_of_objects.append(item.to_dict())
                obj[name_] = jsonutils.dumps(list_of_objects)
            elif isinstance(field, obj_fields.DictOfObjectsField):
                dict_of_objects = {}
                for key, item in getattr(self, name).items():
                    dict_of_objects[key] = item.to_dict()
                obj[name_] = jsonutils.dumps(dict_of_objects)
            elif isinstance(field, obj_fields.DateTimeField):
                obj[name_] = getattr(self, name)
            else:
                obj[name_] = field.to_primitive(self, name,
                                                getattr(self, name))
        return obj


TackerObjectDictCompat = ovoo_base.VersionedObjectDictCompat



# Copyright (c) 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils
from oslo_utils import uuidutils
import sqlalchemy as sa

from tacker.db import types


class HasTenant(object):
    """Tenant mixin, add to subclasses that have a tenant."""

    tenant_id = sa.Column(sa.String(64), nullable=False)


class HasId(object):
    """id mixin, add to subclasses that have an id."""

    id = sa.Column(types.Uuid,
                   primary_key=True,
                   default=uuidutils.generate_uuid)


class Audit(object):
    """Helps to add time stamp for create, update and delete actions. """

    created_at = sa.Column(sa.DateTime,
                           default=lambda: timeutils.utcnow())
    updated_at = sa.Column(sa.DateTime)
    deleted_at = sa.Column(sa.DateTime)
